# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from dataclasses import dataclass
from dataclasses import field
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import openvino as ov
from openvino._pyopenvino.op import Parameter
from openvino._pyopenvino.properties.hint import inference_precision
from openvino.runtime import Node
from openvino.runtime import opset13 as opset

from nncf.common.utils.backend import is_openvino_at_least
from nncf.common.utils.caching import ResultsCache
from nncf.common.utils.caching import cache_results
from nncf.common.utils.cpu_info import is_lnl_cpu
from nncf.common.utils.helpers import set_env_variable
from nncf.openvino.graph.node_utils import convert_op
from nncf.openvino.graph.node_utils import non_convertable_divide_op
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.tensor import Tensor
from nncf.tensor import TensorDataType
from nncf.tensor.functions.openvino_numeric import DTYPE_MAP as DTYPE_MAP_OV

TensorList = List[Tensor]
ModelCallable = Callable[[TensorList], TensorList]
ReductionAxes = Union[int, Tuple[int, ...]]


OV_MODEL_CACHE = ResultsCache()


@dataclass
class OVModelParameters:
    """
    A class to hold parameters for building and inferring an OpenVINO model.

    :param input_dtypes: Optional dictionary mapping input names to their data types.
    :param output_dtypes: Optional dictionary mapping output names to their data types.
    :param dynamic_shapes: Whether to use dynamic shapes for the model. When dynamic shapes are used and
        models are cached, it allows to save on the number of models stored in a model cache.
    :param release_memory: Whether to release memory after every inference. If memory is released, it will be
        reallocated during every inference, reducing performance to some extent.
    :param share_inputs: Whether to share input tensors. Avoids cloning inputs for inference.
    :param share_outputs: Whether to share output tensors. Avoids cloning outputs after the inference.
    :param return_ov_tensors: Whether to return results as OpenVINO tensors or NumPy arrays.
    :param convertable_division: Whether to use convertable division for division operations. If True, division a/b
        will be transformed at runtime to a*(1/b).
    """

    input_dtypes: Dict[str, TensorDataType] = field(default_factory=dict)
    output_dtypes: Dict[str, TensorDataType] = field(default_factory=dict)
    dynamic_shapes: bool = True
    release_memory: bool = True
    share_inputs: bool = True
    share_outputs: bool = True
    return_ov_tensors: bool = False
    convertable_division: bool = False

    def __copy__(self):
        return OVModelParameters(
            input_dtypes=self.input_dtypes.copy(),
            output_dtypes=self.output_dtypes.copy(),
            dynamic_shapes=self.dynamic_shapes,
            release_memory=self.release_memory,
            share_inputs=self.share_inputs,
            share_outputs=self.share_outputs,
            return_ov_tensors=self.return_ov_tensors,
            convertable_division=self.convertable_division,
        )

    def __deepcopy__(self, memo):
        return OVModelParameters(
            input_dtypes=copy.deepcopy(self.input_dtypes, memo),
            output_dtypes=copy.deepcopy(self.output_dtypes, memo),
            dynamic_shapes=self.dynamic_shapes,
            release_memory=self.release_memory,
            share_inputs=self.share_inputs,
            share_outputs=self.share_outputs,
            return_ov_tensors=self.return_ov_tensors,
            convertable_division=self.convertable_division,
        )

    def __hash__(self):
        return hash(
            (
                frozenset(self.input_dtypes.items()),
                frozenset(self.output_dtypes.items()),
                self.dynamic_shapes,
                self.release_memory,
                self.share_inputs,
                self.share_outputs,
                self.return_ov_tensors,
                self.convertable_division,
            )
        )

    def clone(self):
        return copy.deepcopy(self)


ModelAsNodes = Tuple[List[Parameter], List[Node], OVModelParameters]


def clear_ov_model_cache():
    OV_MODEL_CACHE.clear()


def _compile_ov_model(model: ov.Model, device_name: str, config: Dict[str, str]) -> ov.CompiledModel:
    if is_lnl_cpu() and not is_openvino_at_least("2025.1"):
        with set_env_variable("DNNL_MAX_CPU_ISA", "AVX2_VNNI"):
            compiled_model = ov.compile_model(model, device_name=device_name, config=config)
    else:
        compiled_model = ov.compile_model(model, device_name=device_name, config=config)

    return compiled_model


def _infer_ov_model(
    ov_model_params: OVModelParameters, compiled_model: ov.CompiledModel, inputs: TensorList
) -> TensorList:
    """
    Run compiled OpenVINO model inference on the given inputs.
    :param ov_model_params: OV model related parameters.
    :param compiled_model: Compiled OpenVINO model.
    :param inputs: Input tensors.
    :return: List of output tensors. Tensor backend is OV if return_ov_tensors is True, else NumPy.
    """
    # Check that input dtypes match the expected dtypes
    for i, inp in enumerate(compiled_model.inputs):
        input_name = inp.any_name
        actual_dtype = inputs[i].dtype
        expected_dtype = ov_model_params.input_dtypes[input_name]
        if actual_dtype != expected_dtype:
            raise ValueError(f"Expected input '{input_name}' to be {expected_dtype}. But found: {actual_dtype}.")

    # Infer the model
    if compiled_model._infer_request is None:
        compiled_model._infer_request = compiled_model.create_infer_request()
    infer_request = compiled_model._infer_request

    inputs = [inp.data for inp in inputs]
    outputs = infer_request.infer(
        inputs, share_inputs=ov_model_params.share_inputs, share_outputs=ov_model_params.share_outputs
    )
    if ov_model_params.return_ov_tensors:
        outputs = [infer_request.get_output_tensor(i) for i in range(len(outputs))]
    else:
        outputs = [outputs[i] for i in range(len(outputs))]
    outputs = [Tensor(it) for it in outputs]

    if ov_model_params.release_memory:
        compiled_model.release_memory()

    return outputs


def _prepare_compression_model_inputs(
    ov_model_params,
    weight_shape: Tuple,
    scale_shape: Optional[Tuple],
    zero_point_shape: Optional[Tuple],
    reduction_axes: Optional[ReductionAxes],
) -> Tuple[Tuple, Optional[Tuple], Optional[Tuple]]:
    """
    Do some input checks and convert static shapes to dynamic shapes if needed.
    """
    if scale_shape is None and zero_point_shape is not None:
        raise Exception("Zero point shape can only be provided if scale shape is provided.")
    if scale_shape is None and reduction_axes is None:
        raise ValueError("Reduction axes must be provided if scale shape is not provided.")

    # Set dynamic shapes if needed
    if ov_model_params.dynamic_shapes:
        weight_shape = (-1,) * len(weight_shape)
        if scale_shape is not None:
            scale_shape = (-1,) * (len(scale_shape) - 1) + (1,)
        if zero_point_shape is not None:
            zero_point_shape = (-1,) * (len(zero_point_shape) - 1) + (1,)

    return weight_shape, scale_shape, zero_point_shape


def get_compress_weight_model(
    ov_model_params: OVModelParameters,
    config: WeightCompressionConfig,
    weight_shape: Tuple,
    scale_shape: Optional[Tuple] = None,
    zero_point_shape: Optional[Tuple] = None,
    reduction_axes: Optional[ReductionAxes] = None,
    return_nodes: Optional[bool] = False,
) -> Union[ModelCallable, ModelAsNodes]:
    """
    Get a model that compresses weights using the given configuration.
    :param ov_model_params: OV model parameters.
    :param config: Compression configuration.
    :param weight_shape: Shape of the weight to compress. Weight is assumed to be already reshaped as needed.
    :param scale_shape: Optional shape of the scale. If not provided, scale will be computed by the OV model.
        Otherwise, it is expected that the scale tensor is given as an input to the model.
    :param zero_point_shape: Optional shape of the zero point tensor. If not provided and the mode is asymmetric,
        zero point will be computed by the OV model. Otherwise, it is expected that the zero point tensor is provided
        as an input.
    :param reduction_axes: Optional axes to reduce the weight tensor. Not needed if scale (and z.p.) are provided as
        inputs.
    :param return_nodes: Whether to return the OV model inputs parameters and results nodes instead of the model
        callable.
    :return: A model callable that compresses weights using the given configuration. Or a model as nodes, if
        `return_nodes` is True.
    """

    weight_shape, scale_shape, zero_point_shape = _prepare_compression_model_inputs(
        ov_model_params, weight_shape, scale_shape, zero_point_shape, reduction_axes
    )

    return _build_compress_model(
        config,
        ov_model_params,
        weight_shape,
        scale_shape,
        zero_point_shape,
        reduction_axes,
        return_nodes=return_nodes,
    )


def get_compress_decompress_weight_model(
    ov_model_params: OVModelParameters,
    config: WeightCompressionConfig,
    weight_shape: Tuple,
    scale_shape: Optional[Tuple] = None,
    zero_point_shape: Optional[Tuple] = None,
    reduction_axes: Optional[ReductionAxes] = None,
    return_compressed_weight: Optional[bool] = False,
) -> ModelCallable:
    """
    Get a model that performs compression and decompression of the given weight.
    :param ov_model_params: OV model parameters.
    :param config: Compression configuration.
    :param weight_shape: Shape of the weight. Weight is assumed to be already reshaped as needed.
    :param scale_shape: Optional shape of the scale. If not provided, scale will be computed by the OV model.
        Otherwise, it is expected that the scale tensor is given as an input to the model.
    :param zero_point_shape: Optional shape of the zero point tensor. If not provided and the mode is asymmetric,
        zero point will be computed by the OV model. Otherwise, it is expected that the zero point is provided as an
        input.
    :param reduction_axes: Optional axes to reduce the weight tensor. Not needed if scale (and z.p.) are provided as
        inputs.
    :param return_compressed_weight: Whether to also return compressed weight, scale, (and zero point) besides the
        decompressed weight.
    :return: A model callable that returns a decompressed weight, and optionally compressed weight, scale,
        (and zero point) if `return_compressed_weight` is True.
    """

    weight_shape, scale_shape, zero_point_shape = _prepare_compression_model_inputs(
        ov_model_params, weight_shape, scale_shape, zero_point_shape, reduction_axes
    )

    return _build_compress_decompress_model(
        config,
        ov_model_params,
        weight_shape,
        scale_shape,
        zero_point_shape,
        reduction_axes,
        return_compressed_weight,
    )


@cache_results(OV_MODEL_CACHE)
def _build_compress_model(
    config: WeightCompressionConfig,
    ov_model_params: OVModelParameters,
    weight_shape: Tuple,
    scale_shape: Optional[Tuple] = None,
    zero_point_shape: Optional[Tuple] = None,
    reduction_axes: Optional[ReductionAxes] = None,
    return_nodes: bool = False,
) -> Union[ModelCallable, ModelAsNodes]:
    is_asym_mode = config.is_asym_mode

    default_input_dtypes = {
        "scale": TensorDataType.float32,
        "zero_point": TensorDataType.int32,
    }
    default_output_dtypes = {
        "compressed_weight": TensorDataType.uint8 if is_asym_mode else TensorDataType.int8,
        "scale": TensorDataType.float32,
        "zero_point": TensorDataType.int32,
    }

    # Update input and output dtypes with the default values
    ov_model_params = copy.deepcopy(ov_model_params)
    ov_model_params.input_dtypes = {**default_input_dtypes, **ov_model_params.input_dtypes}
    ov_model_params.output_dtypes = {**default_output_dtypes, **ov_model_params.output_dtypes}

    if "weight" not in ov_model_params.input_dtypes:
        raise ValueError("Input weight dtype is required!")

    weight_dtype = ov_model_params.input_dtypes["weight"]
    input_scale_dtype = ov_model_params.input_dtypes["scale"]
    input_zero_point_dtype = ov_model_params.input_dtypes["zero_point"]
    compressed_weight_dtype = ov_model_params.output_dtypes["compressed_weight"]
    output_scale_dtype = ov_model_params.output_dtypes["scale"]
    output_zero_point_dtype = ov_model_params.output_dtypes["zero_point"]

    # Validate input dtypes
    valid_weight_dtypes = [TensorDataType.float32, TensorDataType.float16, TensorDataType.bfloat16]
    if weight_dtype not in valid_weight_dtypes:
        raise ValueError(
            f"Weight must be one of the following data types: {valid_weight_dtypes}. But found: {weight_dtype}."
        )
    if scale_shape is not None and input_scale_dtype != TensorDataType.float32:
        raise ValueError(f"Input scale must be of float32 data type. But found: {input_scale_dtype}.")
    if zero_point_shape is not None and input_zero_point_dtype not in [TensorDataType.int32, TensorDataType.float32]:
        raise ValueError(f"Input zero point must be of int32/float32 data type. But found: {input_zero_point_dtype}.")

    # Validate output dtypes
    valid_compressed_weight_dtypes = [
        TensorDataType.float32,
        TensorDataType.int32,
        TensorDataType.int8,
        TensorDataType.uint8,
        TensorDataType.int4,
        TensorDataType.uint4,
    ]
    if compressed_weight_dtype not in valid_compressed_weight_dtypes:
        raise ValueError(
            f"Compressed weight must be one of the following data types: {valid_compressed_weight_dtypes}. "
            f"But found: {compressed_weight_dtype}."
        )
    if scale_shape is None and output_scale_dtype != TensorDataType.float32:
        raise ValueError(f"Output scale must be of float32 data type. But found: {output_scale_dtype}.")
    if is_asym_mode and zero_point_shape is None and output_zero_point_dtype not in valid_compressed_weight_dtypes:
        raise ValueError(
            f"Output zero point must be of one of the following data types: {valid_compressed_weight_dtypes}. "
            f"But found: {output_zero_point_dtype}."
        )

    # Build OV model
    weight = opset.parameter(weight_shape, name="weight", dtype=DTYPE_MAP_OV[weight_dtype])
    ov_parameters = [weight]

    num_bits = config.num_bits
    eps = np.finfo(np.float32).eps
    level_low = 0 if is_asym_mode else -(2 ** (num_bits - 1))
    level_high = 2**num_bits - 1 if is_asym_mode else 2 ** (num_bits - 1) - 1

    divide_op = opset.divide if ov_model_params.convertable_division else non_convertable_divide_op

    min_values = None
    if scale_shape is not None:
        # Scale is given as an input
        scale = opset.parameter(scale_shape, name="scale", dtype=DTYPE_MAP_OV[input_scale_dtype])
        ov_parameters.append(scale)
    else:
        # Compute scale
        if is_asym_mode:
            # [a1, r, a2] -> [a1, 1, a2]
            min_values = opset.reduce_min(weight, reduction_axes=reduction_axes, keep_dims=True)
            max_values = opset.reduce_max(weight, reduction_axes=reduction_axes, keep_dims=True)
            min_values, max_values = opset.convert(min_values, ov.Type.f32), opset.convert(max_values, ov.Type.f32)

            levels = level_high - level_low + 1
            scale = divide_op(max_values - min_values, opset.constant(levels - 1, ov.Type.f32))
            scale = opset.select(opset.less(opset.abs(scale), eps), eps, scale)
        else:
            w_abs_min = opset.abs(opset.reduce_min(weight, reduction_axes=reduction_axes, keep_dims=True))
            w_max = opset.reduce_max(weight, reduction_axes=reduction_axes, keep_dims=True)
            w_abs_min, w_max = opset.convert(w_abs_min, ov.Type.f32), opset.convert(w_max, ov.Type.f32)

            scale = opset.select(opset.greater_equal(w_abs_min, w_max), w_abs_min, opset.negative(w_max))
            scale = divide_op(scale, opset.constant(-level_low, ov.Type.f32))
            scale = opset.select(opset.less(opset.abs(scale), eps), eps, scale)

    zero_point = None
    if zero_point_shape is not None:
        # Zero point is given as an input
        zero_point = opset.parameter(zero_point_shape, name="zero_point", dtype=DTYPE_MAP_OV[input_zero_point_dtype])
        ov_parameters.append(zero_point)
        # Cast to float32 for an addition later
        zero_point = convert_op(zero_point, ov.Type.f32)
    elif is_asym_mode:
        # Compute zero point
        scaled_min_values = divide_op(min_values, scale)
        zero_point = opset.constant(level_low, ov.Type.f32) - opset.round(scaled_min_values)
        zero_point = opset.clamp(zero_point, level_low, level_high)

    weight = convert_op(weight, ov.Type.f32)
    compressed_weight = divide_op(weight, scale)

    if is_asym_mode:
        compressed_weight += zero_point

    compressed_weight = opset.round(compressed_weight)
    compressed_weight = opset.clamp(opset.round(compressed_weight), level_low, level_high)
    compressed_weight = convert_op(compressed_weight, DTYPE_MAP_OV[compressed_weight_dtype])

    ov_results = [compressed_weight]
    if len(ov_parameters) == 1:
        ov_results.append(scale)
        if zero_point is not None:
            zero_point = convert_op(zero_point, DTYPE_MAP_OV[output_zero_point_dtype])
            ov_results.append(zero_point)

    if return_nodes:
        return ov_parameters, ov_results, ov_model_params

    model = ov.Model(ov_results, ov_parameters)
    compiled_model = _compile_ov_model(model, device_name="CPU", config={inference_precision(): ov.Type.f32})

    return partial(_infer_ov_model, ov_model_params, compiled_model)


@cache_results(OV_MODEL_CACHE)
def _build_compress_decompress_model(
    config: WeightCompressionConfig,
    ov_model_params: OVModelParameters,
    weight_shape: Tuple,
    scale_shape: Optional[Tuple] = None,
    zero_point_shape: Optional[Tuple] = None,
    reduction_axes: Optional[ReductionAxes] = None,
    return_compressed_weight: Optional[bool] = False,
) -> ModelCallable:
    default_output_dtypes = {"decompressed_weight": TensorDataType.float32}
    if not return_compressed_weight:
        # If compressed weight is not returned to a user, we can keep it in float32 to avoid additional conversion
        default_output_dtypes["compressed_weight"] = TensorDataType.float32
    ov_model_params = copy.deepcopy(ov_model_params)
    ov_model_params.output_dtypes = {**default_output_dtypes, **ov_model_params.output_dtypes}

    decompressed_weight_dtype = ov_model_params.output_dtypes["decompressed_weight"]
    if decompressed_weight_dtype != TensorDataType.float32:
        raise ValueError(f"Decompressed weight must be of float32 data type. But found: {decompressed_weight_dtype}.")

    # Get compression model as input/result nodes and potentially modified ov model parameters
    ov_parameters, ov_results, ov_model_params = get_compress_weight_model(
        ov_model_params, config, weight_shape, scale_shape, zero_point_shape, reduction_axes, return_nodes=True
    )

    if config.is_asym_mode:
        if len(ov_parameters) == 1:
            # weight -> compressed_weight, scale, zero_point
            compressed_weight, scale, zero_point = ov_results
        else:
            # weight, scale, zero_point -> compressed_weight
            compressed_weight = ov_results[0]
            scale, zero_point = ov_parameters[1:]

        compressed_weight = convert_op(compressed_weight, ov.Type.i32) - convert_op(zero_point, ov.Type.i32)
    else:
        if len(ov_parameters) == 1:
            # weight -> compressed_weight, scale
            compressed_weight, scale = ov_results
        else:
            # weight, scale -> compressed_weight
            compressed_weight = ov_results[0]
            scale = ov_parameters[1]

    decompressed_weight = opset.multiply(scale, convert_op(compressed_weight, ov.Type.f32))

    ov_results = [decompressed_weight] + ov_results if return_compressed_weight else [decompressed_weight]
    model = ov.Model(ov_results, ov_parameters)
    compiled_model = _compile_ov_model(model, device_name="CPU", config={inference_precision(): ov.Type.f32})

    return partial(_infer_ov_model, ov_model_params, compiled_model)


def get_astype_model(ov_model_params: OVModelParameters, input_shape: Tuple) -> ModelCallable:
    """
    Return a model that cast the input of the given shape to the given data type. Especially useful for
    casting from/to data types not supported by NumPy such as bfloat16, uint4 and int4.
    These data types are represented as the following data types in numpy:
        - bfloat16 -> np.float16,
        - uint4 -> uint8,
        - int4 -> int8.
    :param ov_model_params: OV model related parameters.
    :param input_shape: Shape of the tensor to cast.
    :return: A model callable that casts the input tensor to the given data type.
    """
    if ov_model_params.dynamic_shapes:
        input_shape = (-1,) * len(input_shape)
    return _build_astype_model(ov_model_params, input_shape)


@cache_results(OV_MODEL_CACHE)
def _build_astype_model(ov_model_params: OVModelParameters, arg_shape: Tuple) -> ModelCallable:
    input_dtypes = ov_model_params.input_dtypes
    if input_dtypes is None:
        raise ValueError("Input dtypes must be provided.")
    output_dtypes = ov_model_params.output_dtypes
    if output_dtypes is None:
        raise ValueError("Output dtypes must be provided.")
    if "input" not in input_dtypes:
        raise ValueError("Input dtype is required.")
    if "output" not in output_dtypes:
        raise ValueError("Output dtype is required.")

    arg = opset.parameter(arg_shape, dtype=DTYPE_MAP_OV[input_dtypes["input"]], name="input")
    res = opset.convert(arg, DTYPE_MAP_OV[output_dtypes["output"]])
    model = ov.Model([res], [arg])
    compiled_model = _compile_ov_model(model, device_name="CPU", config={inference_precision(): ov.Type.f32})

    return partial(_infer_ov_model, ov_model_params, compiled_model)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Iterable, List, Optional, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.registry import Registry
from nncf.parameters import SensitivityMetric
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_int_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_int_quantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import get_integer_quantization_error
from nncf.tensor import Tensor
from nncf.tensor import functions as fns
from nncf.tensor.definitions import TensorDataType

TModel = TypeVar("TModel")
MIXED_PRECISION_CRITERIA = Registry("mixed_precision_criteria")
THE_LOWEST_SENSITIVITY = 0


class MixedPrecisionCriterion(Algorithm):
    """
    Assigns mixed quantization scheme (e.g. uniform int8 or uniform int4/non-uniform fp4)
    for weights based on some criteria.
    """

    def __init__(self, primary_config: WeightCompressionConfig, ratio: float, subset_size: Optional[int] = None):
        """
        :param primary_config: Configuration on how to compress (quantize) weights to primary precision.
        :param ratio: The ratio between primary and backup precisions (e.g. 0.9 means 90% of layers quantized to NF4
            and the rest to INT8_ASYM).
        :param subset_size: Size of dataset subset for statistics.
        """
        self._primary_config = primary_config
        self._ratio = ratio
        self._subset_size = subset_size
        self._algorithm_key = f"MPC_{hash(self)}"
        self._backend_entity = None

    @abstractmethod
    def _calc_sensitivity(
        self,
        model: TModel,
        graph: NNCFGraph,
        weight_params: List[WeightCompressionParameters],
        statistic_points: Optional[StatisticPointsContainer] = None,
    ) -> List[float]:
        """
        Calculates sensitivity of each layer according to a criterion.

        :return: List of values per node to be quantized.
        """

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
        weight_params: List[WeightCompressionParameters] = None,
    ) -> None:
        """
        Assigns quantization precision based on computed layers' sensitivities, ratio of parameters.
        """
        self._set_backend_entity(model)

        scores = self._calc_sensitivity(model, graph, weight_params, statistic_points)
        num_all_weights = sum(wp.num_weights for wp in weight_params)

        indexes_of_layers_in_ascending_order_of_scores = [
            i[0] for i in sorted(enumerate(scores), reverse=False, key=lambda x: x[1])
        ]
        num_weights_in_4bit = 0
        for index in indexes_of_layers_in_ascending_order_of_scores:
            weight_param = weight_params[index]
            current_ratio = (num_weights_in_4bit + weight_param.num_weights) / num_all_weights
            if current_ratio >= self._ratio:
                break
            weight_param.compression_config = self._primary_config
            num_weights_in_4bit += weight_param.num_weights

    @abstractmethod
    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """

    @abstractmethod
    def get_statistic_points(
        self,
        model: TModel,
        graph: NNCFGraph,
        nodes_and_port_ids: Iterable[Tuple[NNCFNode, int]],
    ) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: Model for statistics collection.
        :param graph: Model graph.
        :param nodes_and_port_ids: Nodes and port ids for which statistics should be collected.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.WEIGHT_QUANTIZATION_ERROR)
class DataFreeCriterion(MixedPrecisionCriterion):
    """
    A baseline mixed precision criterion that is based on quantization noise of weights only.
    """

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO, BackendType.TORCH, BackendType.TORCH_FX]

    def _set_backend_entity(self, model: TModel) -> None:
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model)
        elif model_backend == BackendType.TORCH:
            from nncf.quantization.algorithms.weight_compression.torch_backend import PTWeightCompressionAlgoBackend

            self._backend_entity = PTWeightCompressionAlgoBackend()
        elif model_backend == BackendType.TORCH_FX:
            from nncf.quantization.algorithms.weight_compression.torch_fx_backend import FXWeightCompressionAlgoBackend

            self._backend_entity = FXWeightCompressionAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def _calc_weight_sensitivity(
        self,
        weight_param: WeightCompressionParameters,
        model: TModel,
        graph: NNCFGraph,
    ) -> float:
        weight = self._backend_entity.get_weight(
            weight_param.node_with_weight, weight_param.weight_port_id, model, graph
        )
        backup_config = WeightCompressionConfig()
        reduction_axes = weight_param.reduction_axes
        int_error = get_integer_quantization_error(weight, reduction_axes, backup_config)
        eps = fns.finfo(weight).eps
        return 1 / (int_error + eps)

    def _calc_score_per_node(
        self,
        weight_param: WeightCompressionParameters,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
    ) -> float:
        weight_score = self._calc_weight_sensitivity(weight_param, model, graph)
        return weight_score

    def _calc_sensitivity(
        self,
        model: TModel,
        graph: NNCFGraph,
        weight_params: List[WeightCompressionParameters],
        statistic_points: Optional[StatisticPointsContainer] = None,
    ) -> List[float]:
        scores = []
        for weight_param in track(weight_params, description="Mixed-Precision assignment"):
            scores.append(self._calc_score_per_node(weight_param, model, graph, statistic_points))
        return scores

    def get_statistic_points(
        self,
        model: TModel,
        graph: NNCFGraph,
        nodes_and_port_ids: Iterable[Tuple[NNCFNode, int]],
    ) -> StatisticPointsContainer:
        raise RuntimeError("No statistics collection intended for data-free mixed precision criterion")


class DataBasedCriterion(DataFreeCriterion, ABC):
    """
    Data-based mixed precision criterion that takes into account outliers in the input statistics.
    Expecting statistics of the following shape: [hidden_dim]
    """

    STAT_KEY = None

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVMixedPrecisionAlgoBackend

            self._backend_entity = OVMixedPrecisionAlgoBackend(model)
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def _calc_activation_sensitivity(
        self,
        weight_param: WeightCompressionParameters,
        graph: NNCFGraph,
        statistic_points: StatisticPointsContainer,
    ) -> float:
        stats = self._get_statistics_for_node(statistic_points, weight_param.node_with_weight, graph, self.STAT_KEY)
        return stats[0].item()

    def _calc_score_per_node(
        self,
        weight_param: WeightCompressionParameters,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
    ):
        """
        NOTE: Data-based criteria for assigning 4-bit/8-bit precisions are valid for Matmul operations only.
        However, in some cases it can be beneficial to quantize Gather layers to 4-bit.
        Since there's no data-aware estimation of sensitivity in these layers, they receive the lowest sensitivity.
        It allows assigning Gather operation 4-bit in the first place.
        """
        if weight_param.node_with_weight.metatype in self._backend_entity.embedding_metatypes:
            return THE_LOWEST_SENSITIVITY
        weight_score = self._calc_weight_sensitivity(weight_param, model, graph)
        activation_score = self._calc_activation_sensitivity(weight_param, graph, statistic_points)
        return weight_score * activation_score

    def get_statistic_points(
        self,
        model: TModel,
        graph: NNCFGraph,
        nodes_and_port_ids: Iterable[Tuple[NNCFNode, int]],
    ) -> StatisticPointsContainer:
        self._set_backend_entity(model)

        statistic_container = StatisticPointsContainer()
        for act_node, output_port_id in nodes_and_port_ids:
            n_dims = len(graph.get_output_edges_by_port_id(act_node, output_port_id)[0].tensor_shape)
            if n_dims < 2:
                raise RuntimeError(
                    f"Data-aware mixed precision criteria are not supported for MatMuls with 1D inputs. "
                    f"Node: {act_node.node_name}, number of dimensions: {n_dims}."
                )
            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, act_node.node_name, port_id=output_port_id
            )
            stat_collector = self._get_statistic_collector()
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )

        return statistic_container

    @abstractmethod
    def _get_statistic_collector(self):
        """
        Get statistic collector
        """

    def _get_activation_node_and_port(self, node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[NNCFNode, int]:
        """
        This method returns the activation layer and corresponding port id for the node.

        :param node: NNCFGraph node for which the activation is sought.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Tuple with the activation node and port id.
        """
        activation_port = self._backend_entity.get_activation_port_id(node, nncf_graph)
        activation_edge = nncf_graph.get_input_edge_by_port_id(node, activation_port)
        activation_node = activation_edge.from_node
        port_id = activation_edge.output_port_id
        return activation_node, port_id

    def _get_statistics_for_node(
        self, statistic_points: StatisticPointsContainer, node: NNCFNode, nncf_graph: NNCFGraph, stat_key: str
    ) -> List[Tensor]:
        act_node, output_port_id = self._get_activation_node_and_port(node, nncf_graph)

        def input_filter_func(point):
            # For the floating-point statistics collected in POST_LAYER style,
            # we also need to determine the output port id.
            # For the cases when the layer has more than one (0) output port.
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
                and point.target_point.port_id == output_port_id
            )

        stats = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            act_node.node_name, input_filter_func, self._algorithm_key
        ):
            statistics = tensor_collector.get_statistics()
            for data in statistics.get_data().values():
                if isinstance(data, Tensor):
                    stats.append(data)
                else:
                    stats.extend(data)
        return stats


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.HESSIAN_INPUT_ACTIVATION)
class HAWQCriterion(DataBasedCriterion):
    """
    Calculates the average Hessian trace of weights with respect to the layer-wise quantization error
    multiplied by L2 norm of 8-bit quantization noise.
    """

    STAT_KEY = SensitivityMetric.HESSIAN_INPUT_ACTIVATION.value

    def _calc_weight_sensitivity(
        self,
        weight_param: WeightCompressionParameters,
        model: TModel,
        graph: NNCFGraph,
    ) -> float:
        weight = self._backend_entity.get_weight(
            weight_param.node_with_weight, weight_param.weight_port_id, model, graph
        )
        backup_config = WeightCompressionConfig()
        reduction_axes = weight_param.reduction_axes

        orig_shape = weight.shape

        if weight.dtype != TensorDataType.float32:
            weight = weight.astype(TensorDataType.float32)

        compressed_weights, scale, zero_point = do_int_quantization(weight, backup_config, reduction_axes)
        decompressed_weight = do_int_dequantization(compressed_weights, scale, zero_point)
        decompressed_weight = decompressed_weight.reshape(orig_shape)
        return fns.linalg.norm(decompressed_weight - weight, ord="fro").item()

    def _get_statistic_collector(self):
        return self._backend_entity.hawq_statistic_collector(self._subset_size)


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MEAN_ACTIVATION_VARIANCE)
class MeanVarianceCriterion(DataBasedCriterion):
    """
    The mean variance of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    STAT_KEY = SensitivityMetric.MEAN_ACTIVATION_VARIANCE.value

    def _get_statistic_collector(self):
        # Reducing across the second-last dimension, assuming it is the sequence length dimension
        return self._backend_entity.mean_variance_statistic_collector(
            reduction_axes=(-2,), subset_size=self._subset_size
        )


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MAX_ACTIVATION_VARIANCE)
class MaxVarianceCriterion(DataBasedCriterion):
    """
    The maximum variance of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    STAT_KEY = SensitivityMetric.MAX_ACTIVATION_VARIANCE.value

    def _get_statistic_collector(self):
        # Reducing across the second-last dimension, assuming it is the sequence length dimension
        return self._backend_entity.max_variance_statistic_collector(
            reduction_axes=(-2,), subset_size=self._subset_size
        )


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MEAN_ACTIVATION_MAGNITUDE)
class MeanMaxCriterion(DataBasedCriterion):
    """
    The mean magnitude of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    STAT_KEY = SensitivityMetric.MEAN_ACTIVATION_MAGNITUDE.value

    def _get_statistic_collector(self):
        # Reducing across the second-last dimension, assuming it is the sequence length dimension
        return self._backend_entity.mean_abs_max_statistic_collector(
            reduction_axes=(-2,), subset_size=self._subset_size
        )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.fx
from torch.ao.quantization.fx.utils import create_getattr_from_value
from torch.ao.quantization.pt2e.utils import _fuse_conv_bn_
from torch.quantization.fake_quantize import FakeQuantize

import nncf
import nncf.torch
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.torch.fx.constant_folding import constant_fold
from nncf.experimental.torch.fx.node_utils import get_graph_node_by_name
from nncf.experimental.torch.fx.node_utils import get_tensor_constant_from_node
from nncf.torch.graph.transformations.commands import PTTargetPoint

TransformationFNType = Callable[[torch.fx.GraphModule], None]

# Copied from torch.fx.Node
BaseArgumentTypes = Union[
    str,
    int,
    float,
    bool,
    complex,
    torch.dtype,
    torch.Tensor,
    torch.device,
    torch.memory_format,
    torch.layout,
    torch._ops.OpOverload,
]
Argument = Optional[Union[Tuple[Any, ...], List[Any], Dict[str, Any], slice, range, torch.fx.Node, BaseArgumentTypes]]

QUANTIZE_NODE_TARGETS = [
    torch.ops.quantized_decomposed.quantize_per_tensor.default,
    torch.ops.quantized_decomposed.quantize_per_tensor.tensor,
    torch.ops.quantized_decomposed.quantize_per_channel.default,
]
DEQUANTIZE_NODE_TARGETS = [
    torch.ops.quantized_decomposed.dequantize_per_tensor.default,
    torch.ops.quantized_decomposed.dequantize_per_tensor.tensor,
    torch.ops.quantized_decomposed.dequantize_per_channel.default,
]
# Map quantize_per_tensor to dequantize_per_tensor, the same for per_channel and vice-versa
QDQ_PAIR = {
    **{
        quantize_node: dequantize_node
        for quantize_node, dequantize_node in zip(QUANTIZE_NODE_TARGETS, DEQUANTIZE_NODE_TARGETS)
    },
    **{
        dequantize_node: quantize_node
        for quantize_node, dequantize_node in zip(QUANTIZE_NODE_TARGETS, DEQUANTIZE_NODE_TARGETS)
    },
}


def _set_new_node_meta(
    new_node: torch.fx.Node,
    prev_nodes: Tuple[Argument, ...],
    target_module: torch.nn.Module,
    model: torch.fx.GraphModule,
):
    """
    Sets correct meta \"val\" value to the new node.

    :param new_node: The new node.
    :param prev_node: Input node of the new node.
        New node expected to have only one input node.
    :param target_module: Module which is being called by the new node.
    """
    vals = []
    for prev_node in prev_nodes:
        val = prev_node
        if isinstance(prev_node, torch.fx.Node):
            val = (
                prev_node.meta["val"]
                if prev_node.op not in ["get_attr"]
                else get_tensor_constant_from_node(prev_node, model).data
            )
            # TODO(anzr299): Support output port id here.
            # Output port 0 is used by default
            val = val[0] if isinstance(val, tuple) else val
        vals.append(val)
    with torch.no_grad():
        new_node.meta["val"] = target_module(*tuple(vals))


def module_insertion_transformation_builder(
    module_to_insert: torch.nn.Module, target_points: List[PTTargetPoint], target_module_name: str
) -> TransformationFNType:
    """
    Returns transformation which inserts given module to a target model
    and calls given module after each target points replacing inputs/outputs
    of the target node.

    :param module_to_insert: Given torch.nn.Module to insert.
    :param target_points: Target points to insert the target module.
    :param target_module_name: Target model attribute name for the module_to_insert.
    :returns: Transformation which which inserts given module to a target model
        and calls given module after each target points.
    """

    def module_insertion_transformation(model: torch.fx.GraphModule):
        module_attr_name = _set_module_to_the_graph_module(model, module_to_insert, target_module_name)
        # Insert call_module nodes to the model
        graph = model.graph
        for idx, target_point in enumerate(target_points):
            new_node = _insert_call_module(graph, target_point, module_attr_name, f"{module_attr_name}_{idx}")
            target_node = get_graph_node_by_name(graph, target_point.target_node_name)

            if target_point.target_type == TargetType.OPERATOR_POST_HOOK:
                _set_new_node_meta(new_node, [target_node], module_to_insert, model)
                with graph.inserting_after(target_node):
                    for user in list(target_node.users):
                        if user is new_node:
                            continue
                        user.replace_input_with(target_node, new_node)

            else:
                prev_node = _get_node_by_input_port_id(target_node, target_point.input_port_id)
                _set_new_node_meta(new_node, [prev_node], module_to_insert, model)
                target_node.replace_input_with(prev_node, new_node)

    return module_insertion_transformation


def leaf_module_insertion_transformation_builder(
    module_to_insert: torch.nn.Module, target_points: List[PTTargetPoint], target_module_name: str
) -> TransformationFNType:
    """
    Returns transformation which inserts given module to a target model
    and calls given module after each target points.

    :param module_to_insert: Given torch.nn.Module to insert.
    :param target_points: Target points to insert the target module.
    :param target_module_name: Target model attribute name for the module_to_insert.
    :returns: Transformation which which inserts given module to a target model
        and calls given module after each target points.
    """

    def leaf_module_insertion_transformation(model: torch.fx.GraphModule):
        module_attr_name = _set_module_to_the_graph_module(model, module_to_insert, target_module_name)
        # Insert call_module nodes to the model
        graph = model.graph
        for idx, target_point in enumerate(target_points):
            _insert_call_module(graph, target_point, module_attr_name, f"{module_attr_name}_{idx}")

    return leaf_module_insertion_transformation


def constant_update_transformation_builder(
    node: NNCFNode, value: torch.Tensor, input_port_id: int = 1
) -> TransformationFNType:
    """
    Return transformation which updates constant of the given node to the given value.

    :param node: Node which requires bias constant update.
    :param value: New value to use as the node constant.
    :param input_port_id: Port Id of the constant.
    :return: Transformation which updates constant of the given node to the given value.
    """

    def constant_update_transformation(model: torch.fx.GraphModule):
        constant_update_fn(model, get_graph_node_by_name(model.graph, node.node_name), value, input_port_id)

    return constant_update_transformation


def constant_update_fn(
    model: torch.fx.GraphModule,
    node: torch.fx.Node,
    value: torch.Tensor,
    input_port_id: int = 1,
    updated_node_name: Optional[str] = None,
):
    """
    Updates constant of given node on the given input port id with given value.

    :param model: Target torch GraphModule.
    :param node: Given graph node.
    :param value: New value to use as the node constant.
    :param input_port_id: Target constant input port id.
    :param updated_node_name: Name of the constant node after updating. Default is `nodename` + `_updated_constant`.
    """
    graph = model.graph
    old_const = _get_node_by_input_port_id(node, input_port_id)

    if old_const.op != "get_attr":
        raise nncf.InternalError(
            f"Constant on input port {input_port_id} for {node} is expected," f" but node {old_const} is present."
        )

    node_name = updated_node_name if updated_node_name else old_const.name + "_updated_constant"
    # Update metadata of the new constant node.
    consumer_nodes = list(old_const.users)
    # This list of consumer nodes is topologically sorted
    # To ensure the updated node has the right order,
    # we insert constant node before the node placed at the highest order in topological order.
    sorted_consumer_nodes = [node for node in graph.nodes if node in consumer_nodes]

    with graph.inserting_before(sorted_consumer_nodes[0]):
        new_const = create_getattr_from_value(model, graph, node_name, value)

    old_const.replace_all_uses_with(new_const, propagate_meta=True)
    graph.eliminate_dead_code()


def qdq_insertion_transformation_builder(
    quantizer: FakeQuantize, target_points: List[PTTargetPoint]
) -> TransformationFNType:
    """
    Returns transformation which inserts quantize-dequantize operations with parameters
    inherited from the given quantizer to each given target point.

    :param quantizer: Quantizer module to inherit quantization parameters from.
    :param target_points: List of target point used to insert quantize-dequantize pairs.
    :return: Transformation which inserts quantize-dequantize operations with parameters
        inherited from the given quantizer to each given target point.
    """

    def qdq_insertion_transformation(model: torch.fx.GraphModule):
        if any(tp.target_type != TargetType.OPERATION_WITH_WEIGHTS for tp in target_points) and len(target_points) > 1:
            raise nncf.InternalError(
                "Insertion of shared qdq pair for the weights is not supported."
                " Please use non shared qdq pairs for the weights quantization."
            )
        for target_point in target_points:
            insert_one_qdq(model, target_point, quantizer)

    return qdq_insertion_transformation


def node_removal_transformation_builder(node: NNCFNode, input_port_id: int) -> TransformationFNType:
    """
    Returns transformation which removes target node from the model and connects
    target node previous node on the given input port id with all target node outputs.

    :param node: Target node to remove.
    :param input_port_id: Input port id which points to input node which should be connected
        to the target node outputs.
    :return: Transformation which removes target node from the model and connects
        target node previous node on the given input port id with all target node outputs.
    """

    def node_removal_transformation(model: torch.fx.GraphModule):
        target_node = get_graph_node_by_name(model.graph, node.node_name)
        input_node = target_node.all_input_nodes[input_port_id]
        for user in list(target_node.users):
            user.replace_input_with(target_node, input_node)
        model.graph.eliminate_dead_code()

    return node_removal_transformation


def output_insertion_transformation_builder(target_point: PTTargetPoint) -> TransformationFNType:
    """
    Returns transformation which inserts clone operation on the given target point
    and extend the model outputs with the inserted cloned value.

    :param target_point: Target point to insert clone and extend the model outputs.
    :return: Transformation which inserts clone operation on the given target point
        and extend the model outputs with the inserted cloned value.
    """

    def output_insertion_transformation(model: torch.fx.GraphModule):
        graph = model.graph
        target_node = get_graph_node_by_name(graph, target_point.target_node_name)
        input_node = get_input_node(target_point, target_node)

        # Clone node output to safe it from inplace operations affects
        with graph.inserting_after(input_node):
            cloned_input = graph.create_node(
                "call_function",
                torch.ops.aten.clone.default,
                (input_node,),
                name=input_node.name + "_cloned",
            )
        cloned_input.meta["val"] = copy(input_node.meta.get("val"))

        # Update args of the output node as one output could be present in the model
        # TODO(dlaykhov) Support case when there are no outputs in the input model.
        output_nodes = [node for node in model.graph.nodes if node.op == "output"]
        assert len(output_nodes) == 1
        output_node = output_nodes[0]

        args = output_node.args
        assert len(args) == 1
        if isinstance(args[0], torch.fx.Node):
            args = (args,)
        args = tuple(args[0]) + (cloned_input,)
        output_node.args = (args,)

    return output_insertion_transformation


def insert_one_qdq(model: torch.fx.GraphModule, target_point: PTTargetPoint, quantizer: FakeQuantize):
    """
    Inserts quantize-dequantize after the target node to the target model.

    :param model: Target model.
    :param target_node: Target node, quantizer-dequantizer pair is inserted just after the
        target node.
    :param quantizer: Quantizer module to inherit quantization parameters from.
    """

    # Copied from torch.ao.quantization.quantize_pt2e.convert_pt2e
    # 1. extract information for inserting q/dq node from activation_post_process
    node_type = "call_function"
    quantize_op: Optional[Callable] = None

    dtype = torch.int8 if quantizer.quant_min < 0 else torch.uint8
    if quantizer.is_per_channel:
        qparams = {
            "_scale_": quantizer.scale,
            "_zero_point_": quantizer.zero_point,
            "_axis_": quantizer.ch_axis,
            "_quant_min_": quantizer.quant_min,
            "_quant_max_": quantizer.quant_max,
            "_dtype_": dtype,
        }
        quantize_op = torch.ops.quantized_decomposed.quantize_per_channel.default
        dequantize_op = torch.ops.quantized_decomposed.dequantize_per_channel.default
    else:
        qparams = {
            "_scale_": float(quantizer.scale),
            "_zero_point_": int(quantizer.zero_point),
            "_quant_min_": quantizer.quant_min,
            "_quant_max_": quantizer.quant_max,
            "_dtype_": dtype,
        }
        quantize_op = torch.ops.quantized_decomposed.quantize_per_tensor.default
        dequantize_op = torch.ops.quantized_decomposed.dequantize_per_tensor.default

    # 2. replace activation_post_process node with quantize and dequantize
    graph = model.graph
    target_node = get_graph_node_by_name(graph, target_point.target_node_name)
    # TODO(dlyakhov): use metatype to get correct input_port_id
    # Do not quantize already quantized nodes
    # inserting_before handle only order in the graph generated code.
    # so, inserting quantize-dequantize and all constant nodes before the usage of the nodes
    with graph.inserting_before(target_node):
        quantize_op_inputs = [target_node]
        for key, value_or_node in qparams.items():
            # TODO(dlyakhov): we can add the information of whether a value needs to
            # be registered as an attribute in qparams dict itself
            if key in ["_scale_", "_zero_point_"] and (not isinstance(value_or_node, (float, int))):
                # For scale and zero_point values we register them as buffers in the root module.
                # However, note that when the values are not tensors, as in the case of
                # per_tensor quantization, they will be treated as literals.
                # However, registering them as a node seems to cause issue with dynamo
                # tracing where it may consider tensor overload as opposed to default.
                # With extra check of scale and zero_point being scalar, it makes
                # sure that the default overload can be used.
                # TODO(dlaykhov): maybe need more complex attr name here
                qparam_node = create_getattr_from_value(model, graph, target_node.name + key, value_or_node)
                quantize_op_inputs.append(qparam_node)
            else:
                # for qparams that are not scale/zero_point (like axis, dtype) we store
                # them as literals in the graph.
                quantize_op_inputs.append(value_or_node)

    input_node = get_input_node(target_point, target_node)
    quantize_op_inputs[0] = input_node
    meta_val = input_node.meta.get("val")

    ctx_manager = get_ctx_manager(graph, target_point)
    with ctx_manager(target_node):
        quantized_node = graph.create_node(node_type, quantize_op, tuple(quantize_op_inputs), {})
    quantized_node.meta["val"] = copy(meta_val)

    # use the same qparams from quantize op
    dq_inputs = [quantized_node] + quantize_op_inputs[1:]
    if target_point.target_type == TargetType.OPERATOR_POST_HOOK:
        user_dq_nodes = []
        with graph.inserting_after(quantized_node):
            for user in target_node.users:
                if user is quantized_node:
                    continue
                dq_node = graph.call_function(dequantize_op, tuple(dq_inputs), {})
                dq_node.meta["val"] = copy(meta_val)
                user_dq_nodes.append((user, dq_node))

        for user, dq_node in user_dq_nodes:
            user.replace_input_with(target_node, dq_node)
    elif target_point.target_type in [TargetType.OPERATOR_PRE_HOOK, TargetType.OPERATION_WITH_WEIGHTS]:
        with graph.inserting_after(quantized_node):
            dq_node = graph.call_function(dequantize_op, tuple(dq_inputs), {})
            dq_node.meta["val"] = copy(meta_val)

        target_node.replace_input_with(input_node, dq_node)
    else:
        raise nncf.InternalError(f"Unexpected target type: {target_point.target_type}")


def _insert_call_module(
    graph: torch.fx.Graph, target_point: PTTargetPoint, module_attr_name: str, graph_node_name: str
):
    """
    Inserts module call node to the graph after/before the target node using the ctx_manager.

    :param graph: Graph to insert module call node.
    :param target_node: Target node, module call node is being inserted just after the target node.
    :param module_attr_name: The name of the graph attribute which keeps the target module.
    :param graph_node_name: Target name for module call node.
    :return: Inserted module call node.
    """
    target_node = get_graph_node_by_name(graph, target_point.target_node_name)
    input_node = get_input_node(target_point, target_node)
    ctx_manager = get_ctx_manager(graph, target_point)
    with ctx_manager(target_node):
        return graph.create_node("call_module", module_attr_name, (input_node,), {}, name=graph_node_name)


def get_input_node(target_point: PTTargetPoint, target_node: torch.fx.Node) -> torch.fx.Node:
    """
    Returns an input node according to the given target point.

    :param target_point: Given target point.
    :param target_node: The target node of the given target point.
    :return: An input node according to the given target point.
    """
    target_type = target_point.target_type
    if target_type not in [
        TargetType.OPERATOR_PRE_HOOK,
        TargetType.OPERATOR_POST_HOOK,
        TargetType.OPERATION_WITH_WEIGHTS,
    ]:
        raise nncf.InternalError(f"Unexpected target type: {target_type}")
    if target_type == TargetType.OPERATOR_POST_HOOK:
        return target_node

    return _get_node_by_input_port_id(target_node, target_point.input_port_id)


def _get_node_by_input_port_id(node: torch.fx.Node, input_port_id: int) -> torch.fx.Node:
    """
    Retrieves an input node from the given node and the input port id.

    :param node: Given input node.
    :param input_port_id: Given input port id.
    :return: An input node from the given node and the input port id.
    """
    if node.target == torch.ops.aten.cat.default:
        return node.args[0][input_port_id]
    return node.args[input_port_id]


def get_ctx_manager(graph: torch.fx.Graph, target_point: PTTargetPoint) -> Callable:
    """
    Return insertion context manager according to the given target point.
    An insertion context manager sets the point at which create_node and
    companion methods will insert into the torch.fx.Graph.

    :param graph: torch.fx.Graph instance.
    :param target_point: Given target point.
    :return: Insertion context manager according to the given target point.
    """
    if target_point.target_type not in [
        TargetType.OPERATOR_PRE_HOOK,
        TargetType.OPERATOR_POST_HOOK,
        TargetType.OPERATION_WITH_WEIGHTS,
    ]:
        raise nncf.InternalError(f"Unexpected target type: {target_point.target_type}")

    if target_point.target_type == TargetType.OPERATOR_POST_HOOK:
        return graph.inserting_after
    return graph.inserting_before


def _set_module_to_the_graph_module(
    model: torch.fx.GraphModule,
    module_to_insert: torch.nn.Module,
    module_name_in_model: str,
) -> str:
    """
    Sets given module to the given torch.fx.GraphModule with unique name.

    :param graph: Target torch.fx.Graph.
    :param module_to_insert: Module to insert to the target graph.
    :param module_name_in_model: Target model attribute name for the module_to_insert.
    :return: A graph module attribute name which keep given module.
    """
    assert not hasattr(model, module_name_in_model)
    setattr(model, module_name_in_model, module_to_insert)
    return module_name_in_model


def _is_supported_batch_norm_for_training(node: torch.fx.Node):
    """
    Return True if the given node refers to an aten batch norm op QAT supports.
    """
    supported_ops = [
        torch.ops.aten.batch_norm.default,
        torch.ops.aten._native_batch_norm_legit.default,
        torch.ops.aten.cudnn_batch_norm.default,
        torch.ops.aten.miopen_batch_norm.default,
    ]
    return node.target in supported_ops


def _get_pattern_replacement_per_channel() -> (
    Tuple[Callable[[torch.Tensor, torch.Tensor, torch.Tensor, int, int, int, torch.dtype], torch.Tensor]]
):
    """
    Returns the patter and replacement function for the subgraph rewriter to
    match and replace for per_tensor quantization
    """

    def pattern_per_channel(weight, scale, zero_point, axis, low, high, dtype):
        quantized = torch.ops.quantized_decomposed.quantize_per_channel.default(
            weight, scale, zero_point, axis, low, high, dtype
        )
        dequantized = torch.ops.quantized_decomposed.dequantize_per_channel.default(
            quantized, scale, zero_point, axis, low, high, dtype
        )
        return dequantized

    def replacement_graph_per_channel(weight, scale, zero_point, axis, low, high, dtype):
        return torch.ops.aten.sub.Tensor(torch.ops.aten.mul.Tensor(weight, scale), zero_point)

    return pattern_per_channel, replacement_graph_per_channel


def _get_pattern_replacement_per_tensor() -> (
    Tuple[Callable[[torch.Tensor, torch.Tensor, torch.Tensor, int, int, torch.dtype], torch.Tensor]]
):
    """
    Returns the patter and replacement function for the subgraph rewriter to
    match and replace for per_tensor quantization
    """

    def pattern_per_tensor(weight, scale, zero_point, low, high, dtype):
        quantized = torch.ops.quantized_decomposed.quantize_per_tensor.default(
            weight, scale, zero_point, low, high, dtype
        )
        dequantized = torch.ops.quantized_decomposed.dequantize_per_tensor.default(
            quantized, scale, zero_point, low, high, dtype
        )
        return dequantized

    def replacement_graph_per_tensor(weight, scale, zero_point, low, high, dtype):
        return torch.ops.aten.sub.Tensor(torch.ops.aten.mul.Tensor(weight, scale), zero_point)

    return pattern_per_tensor, replacement_graph_per_tensor


def _set_meta_for_matches(model: torch.fx.GraphModule, matches: torch.fx.subgraph_rewriter.ReplacedPatterns) -> None:
    """
    Sets the meta for the nodes which were matched by the subgraph_rewriter pattern.

    :param model: Model whose nodes' meta are to be set.
    :param matches: matches returned by the subgraph_rewriter
    """
    for match in matches:
        mul_node = match.replacements[0]
        sub_node = match.replacements[1]
        _set_new_node_meta(mul_node, mul_node.args, torch.mul, model)
        _set_new_node_meta(sub_node, sub_node.args, torch.sub, model)


def _get_node_inputs(node: torch.fx.Node, model: torch.fx.GraphModule) -> Optional[Tuple[Union[torch.Tensor, int]]]:
    """
    Gets the inputs for the Quantize node which quantize the weights. Otherwise returns None.

    :param node: Node to get the inputs from
    :param model: Model to which the node belongs
    """
    args = []
    for arg in node.args:
        if isinstance(arg, torch.fx.Node):
            if arg.op != "get_attr":
                return None
            arg = get_tensor_constant_from_node(arg, model)
        args.append(arg)
    return tuple(args)


def _get_value(
    arg: Optional[Union[torch.fx.Node, float, int]], model: torch.fx.GraphModule
) -> Union[torch.nn.Parameter, float, int]:
    """
    Retrieves value from the given argument. It can be either torch.fx.Node or float/int value.

    :param arg: Given arg to retrieve value.
    :param model: torch.fx.GraphModule instance.
    :return: value from the given argument.
    """
    if isinstance(arg, torch.fx.Node):
        return get_tensor_constant_from_node(arg, model)
    return arg


def _compress_qdq_constant_transformation(model: torch.fx.GraphModule, matches) -> None:
    """
    Change the FP32 weight value to Int8 and also reshape the scale for per_channel_quantization.

    :param: model: Model to apply transformations to.
    """

    for match in matches:
        mul_node = match.replacements[0]
        sub_node = match.replacements[1]
        nodes_map = {node.name: match.nodes_map[node] for node in match.nodes_map}

        weight_node = _get_value(nodes_map["weight"], model)
        scale_node = _get_value(nodes_map["scale"], model)
        zp_node = _get_value(nodes_map["zero_point"], model)
        axis = _get_value(nodes_map.get("axis"), model)
        port_id = 0
        if axis is not None:
            result = torch.ops.quantized_decomposed.quantize_per_channel.default(
                weight_node, scale_node, zp_node, axis, -128, 127, torch.int8
            )
            _reshape_scale_zp(model, sub_node, mul_node, weight_node, scale_node, zp_node, axis)
        else:
            result = torch.ops.quantized_decomposed.quantize_per_tensor.default(
                weight_node, scale_node, zp_node, -128, 127, torch.int8
            )
        constant_update_fn(model, mul_node, result, port_id, updated_node_name="compressed_weight_updated_constant")


def _reshape_scale_zp(
    model: torch.fx.GraphModule,
    sub_node: torch.fx.Node,
    mul_node: torch.fx.Node,
    weight: torch.Tensor,
    scale: Union[torch.Tensor, float],
    zp: Union[torch.Tensor, float, int],
    axis: int,
) -> None:
    """
    Reshape scale and zero point so that it can be multiplied elementwise with the weight for per channel quantization.

    :param model: Model to apply transformations to.
    :param node: quantize node whose scale has to be reshaped
    """
    new_shape = [1] * weight.dim()
    new_shape[axis] = scale.shape[0]
    scale = scale.reshape(new_shape)
    zp = zp.reshape(new_shape)
    constant_update_fn(model, sub_node, zp, 1, updated_node_name="zero_point_updated_constant")
    constant_update_fn(model, mul_node, scale, 1, updated_node_name="scale_updated_constant")


def fq_weights_transformation(model: torch.fx.GraphModule) -> None:
    """
    This function applies a transformation to replace the FP32 weights with Fake Quantized
    FP values to avoid the rounding error when converting to OpenVino model.

    :param model: Model to apply transformations to.
    """
    for node in model.graph.nodes:
        if node.target not in DEQUANTIZE_NODE_TARGETS:
            continue
        port_id = 0
        quantize_node = node.all_input_nodes[0]
        # Case where Quantize Dequantize nodes are not consecutive
        if quantize_node.target != QDQ_PAIR[node.target]:
            continue
        args = _get_node_inputs(quantize_node, model)
        if not args:
            continue
        result = node.target(quantize_node.target(*args), *args[1:])
        constant_update_fn(
            model,
            quantize_node,
            result,
            port_id,
            updated_node_name="compressed_weight_updated_constant",
        )


def compress_post_quantize_transformation(model: torch.fx.GraphModule) -> None:
    """
    Applies transformation to compress the weights to Int8 after the quantization step.
    Starts by removing the Quantize/De-Quantize nodes for weight nodes by matching the pattern
    to be like follows:

    weight scale zp
        \   |    /
         Quantize
            |
        Dequantize

    and replace with the following:
     weight scale
       \     /
       Multiply  Zero Point
          |     /
        Subtract

    and then also updates the weights to int8 values and reshape the scale and zero point.

    :param model: Model to apply transformations to.
    """

    def match_filters(match, original_graph, graph):
        for node in match.nodes_map:
            if node.name == "weight" and match.nodes_map[node].op == "get_attr":
                return True
        return False

    patterns = [_get_pattern_replacement_per_channel(), _get_pattern_replacement_per_tensor()]

    for pattern, replacement in patterns:
        matches = torch.fx.subgraph_rewriter.replace_pattern_with_filters(model, pattern, replacement, [match_filters])
        _compress_qdq_constant_transformation(model, matches)
        _set_meta_for_matches(model, matches)


def apply_quantization_transformations(model: torch.fx.GraphModule) -> None:
    """
    Applies quantization transformations to the model.

    :param model: Model to apply transformations to.
    """
    # BatchNorm operations have 3 output ports,
    # to make it easier for algorithms to work
    # with the target graph BatchNorm operations
    # are being fused
    fold_constant_except_qdq(model)
    _fuse_conv_bn_(model)


def fold_constant_except_qdq(model: torch.fx.GraphModule):
    """
    Performs constant folding avoiding quantize-dequantize pattern.

    :param model: Model to perform constant folding on.
    """

    def constraint_fn(node: torch.fx.Node):
        return node.op != "call_function" or node.target not in QUANTIZE_NODE_TARGETS + DEQUANTIZE_NODE_TARGETS

    constant_fold(model, constraint_fn=constraint_fn)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Optional, Tuple, Union

from nncf.tensor import Tensor
from nncf.tensor.functions.dispatcher import tensor_guard


@functools.singledispatch
@tensor_guard
def norm(
    a: Tensor,
    ord: Optional[Union[str, float, int]] = None,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
) -> Tensor:
    """
    Computes a vector or matrix norm.

    The following norms can be calculated:

    =====  ============================  ==========================
    ord    norm for matrices             norm for vectors
    =====  ============================  ==========================
    None   Frobenius norm                2-norm
    'fro'  Frobenius norm                --
    'nuc'  nuclear norm                  --
    inf    max(sum(abs(x), axis=1))      max(abs(x))
    -inf   min(sum(abs(x), axis=1))      min(abs(x))
    0      --                            sum(x != 0)
    1      max(sum(abs(x), axis=0))      as below
    -1     min(sum(abs(x), axis=0))      as below
    2      2-norm (largest sing. value)  as below
    -2     smallest singular value       as below
    other  --                            sum(abs(x)**ord)**(1./ord)
    =====  ============================  ==========================

    The Frobenius norm is given by [1]_:

        :math:`||A||_F = [\\sum_{i,j} abs(a_{i,j})^2]^{1/2}`

    The nuclear norm is the sum of the singular values.

    Both the Frobenius and nuclear norm orders are only defined for
    matrices and otherwise raise a ValueError.

    :param a: The input tensor.
    :param ord: Order of norm. Default: None.
    :param axis: Axis over which to compute the vector or matrix norm. Default: None.
    :param keepdims: If set to True, the reduced dimensions are retained in the result
        as dimensions with size one. Default: False.
    :return: Norm of the matrix or vector.
    """
    return Tensor(norm(a.data, ord, axis, keepdims))


@functools.singledispatch
@tensor_guard
def cholesky(a: Tensor, upper: bool = False) -> Tensor:
    """
    Computes the Cholesky decomposition of a complex Hermitian or real symmetric
    positive-definite matrix `a`.

    If `upper` is False, then the Cholesky decomposition, `L * L.H` of the matrix `a`
    is returned, where `L` is lower-triangular and .H is the conjugate transpose operator
    If `upper` is True, then the conjugate transpose tensor of the tensor with upper=False
    is returned.

    :param a: The input tensor of size (N, N).
    :param upper: Whether to compute the upper- or lower-triangular Cholesky factorization.
        Default is lower-triangular.
    :return: Upper- or lower-triangular Cholesky factor of `a`.
    """
    return Tensor(cholesky(a.data, upper))


@functools.singledispatch
@tensor_guard
def cholesky_inverse(a: Tensor, upper: bool = False) -> Tensor:
    """
    Computes the inverse of a complex Hermitian or real symmetric positive-definite matrix given
    its Cholesky decomposition.

    :param a: The input tensor of size (N, N) consisting of lower or upper triangular Cholesky
        decompositions of symmetric or Hermitian positive-definite matrices.
    :param upper: The flag that indicates whether the input tensor is lower triangular or
        upper triangular. Default: False.
    :return: The inverse of matrix given its Cholesky decomposition.
    """
    return Tensor(cholesky_inverse(a.data, upper))


@functools.singledispatch
@tensor_guard
def inv(a: Tensor) -> Tensor:
    """
    Computes the inverse of a matrix.

    :param a: The input tensor of shape (*, N, N) where * is zero or more batch dimensions
        consisting of invertible matrices.
    :return: The inverse of the input tensor.
    """
    return Tensor(inv(a.data))


@functools.singledispatch
@tensor_guard
def pinv(a: Tensor) -> Tensor:
    """
    Compute the (Moore-Penrose) pseudo-inverse of a matrix.

    :param a: The input tensor of shape (*, M, N) where * is zero or more batch dimensions.
    :return: The pseudo-inverse of input tensor.
    """
    return Tensor(pinv(a.data))


@functools.singledispatch
@tensor_guard
def lstsq(a: Tensor, b: Tensor, driver: Optional[str] = None) -> Tensor:
    """
    Compute least-squares solution to equation Ax = b.

    :param a: Left-hand side tensor of size (M, N).
    :param b: Right-hand side tensor of size (M,) or (M, K).
    :param driver: name of the LAPACK/MAGMA method to be used for solving the least-squares problem.
    :return: a tensor of size (N,) or (N, K), such that the 2-norm |b - A x| is minimized.
    """
    return Tensor(lstsq(a.data, b.data, driver))


@functools.singledispatch
@tensor_guard
def svd(a: Tensor, full_matrices: Optional[bool] = True) -> Tensor:
    """
    Factorizes the matrix a into two unitary matrices U and Vh, and a 1-D array s of singular values
    (real, non-negative) such that a == U @ S @ Vh, where S is a suitably shaped matrix of zeros with main diagonal s.

    :param a: Tensor of shape (*, M, N) where * is zero or more batch dimensions.
    :param full_matrices: Controls whether to compute the full or reduced SVD, and consequently,
        If True (default), u and vh have the shapes (..., M, M) and (..., N, N), respectively.
        Otherwise, the shapes are (..., M, K) and (..., K, N), respectively, where K = min(M, N).
    :return: a tuple with the following tensors:
        U - Unitary matrix having left singular vectors as columns. Of shape (M, M) or (M, K),
            depending on full_matrices.
        S - The singular values, sorted in non-increasing order. Of shape (K,), with K = min(M, N).
        Vh - Unitary matrix having right singular vectors as rows. Of shape (N, N) or (K, N) depending on full_matrices.
    """
    U, S, Vh = svd(a.data, full_matrices=full_matrices)
    return (Tensor(U), Tensor(S), Tensor(Vh))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import merge_two_types_of_operations
from nncf.torch.graph import operator_metatypes as om

LINEAR_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [
        # Linear
        om.PTLinearMetatype,
        om.PTModuleLinearMetatype,
        # Conv1D
        om.PTConv1dMetatype,
        om.PTDepthwiseConv1dSubtype,
        om.PTModuleConv1dMetatype,
        om.PTModuleDepthwiseConv1dSubtype,
        # Conv2D
        om.PTConv2dMetatype,
        om.PTDepthwiseConv2dSubtype,
        om.PTModuleConv2dMetatype,
        om.PTModuleDepthwiseConv2dSubtype,
        # Conv3D
        om.PTConv3dMetatype,
        om.PTDepthwiseConv3dSubtype,
        om.PTModuleConv3dMetatype,
        om.PTModuleDepthwiseConv3dSubtype,
        # Transposed conv
        om.PTConvTranspose1dMetatype,
        om.PTModuleConvTranspose1dMetatype,
        om.PTConvTranspose2dMetatype,
        om.PTModuleConvTranspose2dMetatype,
        om.PTConvTranspose3dMetatype,
        om.PTModuleConvTranspose3dMetatype,
        # Deform conv
        om.PTDeformConv2dMetatype,
        om.PTModuleDeformConv2dMetatype,
        # MatMul
        om.PTMatMulMetatype,
        # Addmm
        om.PTAddmmMetatype,
    ],
    GraphPattern.LABEL_ATTR: "LINEAR",
}

BATCH_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [om.PTBatchNormMetatype, om.PTModuleBatchNormMetatype],
    GraphPattern.LABEL_ATTR: "BATCH_NORMALIZATION",
}

GROUP_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [om.PTGroupNormMetatype, om.PTModuleGroupNormMetatype],
    GraphPattern.LABEL_ATTR: "GROUP_NORMALIZATION",
}

LAYER_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [om.PTLayerNormMetatype, om.PTModuleLayerNormMetatype],
    GraphPattern.LABEL_ATTR: "LAYER_NORMALIZATION",
}

RELU_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [
        om.PTRELUMetatype,
        om.PTHardTanhMetatype,
    ],
    GraphPattern.LABEL_ATTR: "RELU",
}

NON_RELU_ACTIVATIONS_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [
        om.PTELUMetatype,
        om.PTPRELUMetatype,
        om.PTLeakyRELUMetatype,
        om.PTSigmoidMetatype,
        om.PTGELUMetatype,
        om.PTSILUMetatype,
        om.PTHardSigmoidMetatype,
        om.PTHardSwishMetatype,
    ],
    GraphPattern.LABEL_ATTR: "NON_RELU_ACTIVATIONS",
}

ATOMIC_ACTIVATIONS_OPERATIONS = merge_two_types_of_operations(
    RELU_OPERATIONS, NON_RELU_ACTIVATIONS_OPERATIONS, "ATOMIC_ACTIVATIONS"
)

ARITHMETIC_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [om.PTAddMetatype, om.PTSubMetatype, om.PTMulMetatype, om.PTDivMetatype],
    GraphPattern.LABEL_ATTR: "ARITHMETIC",
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PyTorch-specific structure definitions for passing arguments into certain NNCF calls.
"""
from typing import Any, Callable, Optional, Tuple

import torch
from torch import nn
from torch.nn.modules.loss import _Loss
from torch.utils.data import DataLoader

from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.utils.api_marker import api
from nncf.config.structures import NNCFExtraConfigStruct


@api()
class QuantizationPrecisionInitArgs(NNCFExtraConfigStruct):
    """
    Stores arguments for initialization of quantization's bitwidth.
    Initialization is based on calculating a measure reflecting layers' sensitivity to perturbations. The measure is
    calculated by estimation of average trace of Hessian for modules using the Hutchinson algorithm.

    :param criterion_fn: callable object, that implements calculation of loss by given outputs of the model, targets,
      and loss function. It's not needed when the calculation of loss is just a direct call of the criterion with 2
      arguments: outputs of model and targets. For all other specific cases, the callable object should be provided.
      E.g. for inception-v3, the losses for two outputs of the model are combined with different weight.
    :param criterion: loss function, instance of descendant of `torch.nn.modules.loss._Loss`,
    :param data_loader: 'data_loader' - provides an iterable over the given dataset. Instance of
      nncf.initialization.PTInitializingDataLoader; a regular 'torch.utils.data.DataLoader' may
      also be passed, but only in the simple case when it returns a tuple of (input, target) tensors.
    .. WARNING:: The final quantizer setup of the created compressed model is dependent on the data
      provided by the data_loader. When using PyTorch's DistributedDataParallel with precision
      initialization, make sure that each process in the distributed group receives the same data
      from the data_loader as the other processes, otherwise the create_compressed_model call may
      create different compressed model objects for each distributed process and the distributed training
      will fail.
    :param device: Device to perform initialization at. Either 'cpu', 'cuda', or None (default); if None, will
      use the device of the model's parameters.
    """

    def __init__(
        self,
        criterion_fn: Callable[[Any, Any, _Loss], torch.Tensor],
        criterion: _Loss,
        data_loader: DataLoader,
        device: str = None,
    ):
        self.criterion_fn = criterion_fn
        self.criterion = criterion
        self.data_loader = data_loader
        self.device = device

    @classmethod
    def get_id(cls) -> str:
        return "quantization_precision_init_args"


@api()
class AutoQPrecisionInitArgs(NNCFExtraConfigStruct):
    """
    :param data_loader: 'data_loader' - provides an iterable over the given dataset. Instance of
      nncf.initialization.PTInitializingDataLoader; a regular 'torch.utils.data.DataLoader' may
      also be passed, but only in the simple case when it returns a tuple of (input, target) tensors.
     .. WARNING:: The final quantizer setup of the created compressed model is dependent on the data
      provided by the data_loader. When using PyTorch's DistributedDataParallel with precision
      initialization, make sure that each process in the distributed group receives the same data
      from the data_loader as the other processes, otherwise the create_compressed_model call may
      create different compressed model objects for each distributed process and the distributed training
      will fail.
    """

    def __init__(
        self,
        data_loader: DataLoader,
        eval_fn: Callable[[torch.nn.Module, torch.utils.data.DataLoader], float],
        nncf_config: NNCFConfig,
    ):
        self.data_loader = data_loader
        self.eval_fn = eval_fn
        self.config = nncf_config

    @classmethod
    def get_id(cls) -> str:
        return "autoq_precision_init_args"


@api()
class LeGRInitArgs(NNCFExtraConfigStruct):
    """
    Stores arguments for learning global ranking in pruning algorithm.

    :param train_loader: provides an iterable over the given training (or initialising) dataset.
    :param train_fn: callable for training compressed model. Train model for one epoch or train_steps (if specified) by
      given args: [dataloader, model, optimizer, compression algorithm controller, train_steps number].
    :param val_loader: provides an iterable over the given validation dataset.
    :param val_fn: callable to validate model, calculates pair of validation [acc, loss] by given model and dataloader.
    :param train_optimizer: optional, optimizer for model training.
    :param nncf_config: NNCF config for compression.
    """

    def __init__(
        self,
        train_loader: torch.utils.data.DataLoader,
        train_fn: Callable[
            [
                torch.utils.data.DataLoader,
                torch.nn.Module,
                torch.optim.Optimizer,
                CompressionAlgorithmController,
                Optional[int],
            ],
            type(None),
        ],
        val_loader: torch.utils.data.DataLoader,
        val_fn: Callable[[torch.nn.Module, torch.utils.data.DataLoader], Tuple[float, float]],
        train_optimizer: Optional[torch.optim.Optimizer],
        nncf_config: NNCFConfig,
    ):
        self.train_loader = train_loader
        self.train_steps_fn = train_fn
        self.val_loader = val_loader
        self.val_fn = val_fn
        self.train_optimizer = train_optimizer
        self.config = nncf_config

    @classmethod
    def get_id(cls) -> str:
        return "legr_init_args"


@api()
class DistributedCallbacksArgs(NNCFExtraConfigStruct):
    """
    A pair of callbacks that is needed for distributed training of the model: wrapping model with wrapping_callback for
    distributed training, and after all training steps unwrapping model to the initial not-distributed state with
    unwrapping_callback.

    :param wrapping_callback: Callback that wraps the model for distributed training with any necessary structure (for
      example, torch.nn.DataParallel or any custom class), returns wrapped model ready for distributed training
    :param unwrapping_callback: Callback for unwrapping the model wrapped with wrapping_callback, returns original model
    """

    def __init__(
        self, wrapping_callback: Callable[[nn.Module], nn.Module], unwrapping_callback: Callable[[nn.Module], nn.Module]
    ):
        self.wrap_model = wrapping_callback
        self.unwrap_model = unwrapping_callback

    @classmethod
    def get_id(cls) -> str:
        return "distributed_callbacks_args"


@api()
class ExecutionParameters:
    """
    Parameters that are necessary for distributed training of the model.

    :param cpu_only: whether cpu-only mode is using for training
    :param current_gpu: id of GPU that should be used for training (if only one of all is used)
    """

    def __init__(self, cpu_only: bool, current_gpu: Optional[int]):
        self.cpu_only = cpu_only
        self.current_gpu = current_gpu



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
from typing import Any, Dict, Generic, Optional, Tuple, TypeVar

_K = TypeVar("_K")
_V = TypeVar("_V")


class WeakUnhashableKeyMap(Generic[_K, _V]):
    """
    A dictionary-like object that uses weak references to unhashable objects as keys.

    This class allows the use of unhashable objects as keys in a dictionary by using their `id`.
    It ensures that the entries are removed when the objects are garbage collected.

    :param _data: Internal storage for the key-value pairs,
        where the key is the id of the object and the value is a tuple containing a weak reference
        to the object and the actual value.
    """

    def __init__(self) -> None:
        """Initialize an empty WeakUnhashableKeyMap."""
        self._data: Dict[int, Tuple[weakref.ReferenceType[_K], _V]] = {}

    def __getitem__(self, obj: Any) -> Any:
        """
        Get the value associated with the given object.

        :param obj (Any): The key object.
        :return: The value associated with the object.
        """
        ref_obj, val = self._data[id(obj)]
        if ref_obj() is not obj:
            raise KeyError(obj)
        return val

    def __setitem__(self, obj: _K, value: _V) -> None:
        """
        Set the value for the given object.

        :param obj: The key object.
        :param value: The value to associate with the object.
        """
        key = id(obj)
        ref_obj, _ = self._data.get(key, (None, None))

        if ref_obj is not None and ref_obj() is obj:
            self._data[key] = ref_obj, value
        else:

            def on_destroy(_: Any) -> None:
                del self._data[key]

            self._data[key] = weakref.ref(obj, on_destroy), value

    def get(self, obj: _K, default: Optional[Any] = None) -> Optional[_V]:
        """
        Get the value associated with the given object, or a default value if the object is not in the map.

        :param obj: The key object.
        :param default: The default value to return if the object is not found. Defaults to None.
        :return: The value associated with the object, or the default value if the object is not found.
        """
        key = id(obj)
        try:
            _, val = self._data[key]
            return val
        except KeyError:
            return default



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List, Set

import numpy as np
import tensorflow as tf

from nncf import NNCFConfig
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.scopes import should_consider_scope
from nncf.common.sparsity.schedulers import SPARSITY_SCHEDULERS
from nncf.common.sparsity.schedulers import SparsityScheduler
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.config.schemata.defaults import SPARSITY_LEVEL_SETTING_MODE
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.converter import TFModelConverterFactory
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx
from nncf.tensorflow.sparsity.base_algorithm import SPARSITY_LAYER_METATYPES
from nncf.tensorflow.sparsity.base_algorithm import BaseSparsityController
from nncf.tensorflow.sparsity.collector import TFSparseModelStatisticsCollector
from nncf.tensorflow.sparsity.rb.loss import SparseLoss
from nncf.tensorflow.sparsity.rb.operation import RBSparsifyingWeight


@TF_COMPRESSION_ALGORITHMS.register("rb_sparsity")
class RBSparsityBuilder(TFCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        self.ignored_scopes = self._algo_config.get("ignored_scopes", [])
        self._op_names = []

    def get_transformation_layout(self, model: tf.keras.Model) -> TFTransformationLayout:
        converter = TFModelConverterFactory.create(model)
        nncf_graph = converter.convert()

        check_scopes_in_graph(nncf_graph, self.ignored_scopes, self.target_scopes, self.validate_scopes)

        transformations = TFTransformationLayout()

        processed_shared_layer_names: Set[str] = set()

        for node in nncf_graph.get_all_nodes():
            if node.is_shared():
                target_layer_name, _ = get_original_name_and_instance_idx(node.node_name)
                if target_layer_name in processed_shared_layer_names:
                    continue
                processed_shared_layer_names.add(target_layer_name)

            if not (
                node.metatype in SPARSITY_LAYER_METATYPES
                and should_consider_scope(node.node_name, ignored_scopes=self.ignored_scopes)
            ):
                continue

            _, layer_info = converter.get_layer_info_for_node(node.node_name)
            for weight_def in node.metatype.weight_definitions:
                op_name = self._get_rb_sparsity_operation_name(node.node_name, weight_def.weight_attr_name)
                self._op_names.append(op_name)

                transformations.register(
                    TFInsertionCommand(
                        target_point=TFLayerWeight(layer_info.layer_name, weight_def.weight_attr_name),
                        callable_object=RBSparsifyingWeight(op_name),
                        priority=TransformationPriority.SPARSIFICATION_PRIORITY,
                    )
                )

        return transformations

    def _get_rb_sparsity_operation_name(self, layer_name: str, weight_attr_name: str) -> str:
        return f"{layer_name}_{weight_attr_name}_rb_sparsity_weight"

    def _build_controller(self, model: tf.keras.Model) -> "RBSparsityController":
        """
        Simple implementation of building controller without setting builder state and loading controller's one.
        Should be called once the compressed model target_model is fully constructed.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `RBSparsityController`.
        """

        return RBSparsityController(model, self.config, self._op_names)

    def initialize(self, model: tf.keras.Model) -> None:
        pass


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("tf_rb_sparsity")
class RBSparsityController(BaseSparsityController):
    """
    Controller class for regularization-based (RB) sparsity in TF.
    """

    def __init__(self, target_model, config: NNCFConfig, op_names: List[str]):
        super().__init__(target_model, op_names)
        algo_config = extract_algo_specific_config(config, "rb_sparsity")
        sparsity_init = algo_config.get("sparsity_init", SPARSITY_INIT)
        params = deepcopy(algo_config.get("params", {}))
        params["sparsity_init"] = sparsity_init
        sparsity_level_mode = params.get("sparsity_level_setting_mode", SPARSITY_LEVEL_SETTING_MODE)

        if sparsity_level_mode == "local":
            raise NotImplementedError("RB sparsity algorithm do not support local sparsity loss")

        target_ops = []
        for wrapped_layer, _, op in get_nncf_operations(self.model, self._op_names):
            target_ops.append((op, wrapped_layer.get_operation_weights(op.name)))

        self._loss = SparseLoss(target_ops)
        schedule_type = params.get("schedule", "exponential")

        if schedule_type == "adaptive":
            raise NotImplementedError("RB sparsity algorithm do not support adaptive scheduler")

        scheduler_cls = SPARSITY_SCHEDULERS.get(schedule_type)
        self._scheduler = scheduler_cls(self, params)
        self.set_sparsity_level(sparsity_init)

    @property
    def scheduler(self) -> SparsityScheduler:
        return self._scheduler

    @property
    def loss(self) -> SparseLoss:
        return self._loss

    def set_sparsity_level(self, sparsity_level):
        self._loss.set_target_sparsity_loss(sparsity_level)

    @property
    def current_sparsity_level(self) -> float:
        # TODO: align with torch where it currently shows the sparsity level as reported by loss object.
        #  TF does not seem to have this functionality in its SparseLoss right now.
        return self.scheduler.current_sparsity_level

    def freeze(self):
        self._loss.disable()

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = TFSparseModelStatisticsCollector(self.model, self._op_names)
        model_stats = collector.collect()

        sparse_prob_sum = 0.0
        num_weights = 0
        for wrapped_layer, _, op in get_nncf_operations(self.model, self._op_names):
            operation_weights = wrapped_layer.get_operation_weights(op.name)
            mask = op.get_mask(operation_weights)
            sparse_prob_sum += tf.math.reduce_sum(tf.math.sigmoid(mask)).numpy().item()
            num_weights += np.prod(mask.shape.as_list()).item()
        mean_sparse_prob = 1.0 - (sparse_prob_sum / num_weights)

        target_sparsity_level = self.scheduler.current_sparsity_level

        stats = RBSparsityStatistics(model_stats, target_sparsity_level, mean_sparse_prob)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("rb_sparsity", stats)
        return nncf_stats

    @property
    def compression_rate(self) -> float:
        return self._loss.target_sparsity_rate

    @compression_rate.setter
    def compression_rate(self, compression_rate: float) -> None:
        self.set_sparsity_level(compression_rate)

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()

    def compression_stage(self) -> CompressionStage:
        return None  # Issue-160174



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Structures for passing live Python objects into NNCF algorithms.
"""
from typing import Any, Callable, Optional

from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.utils.api_marker import api


class NNCFExtraConfigStruct:
    """
    This is the class from which all extra structures that define additional
    NNCFConfig arguments inherit.
    """

    @classmethod
    def get_id(cls) -> str:
        raise NotImplementedError


@api()
class QuantizationRangeInitArgs(NNCFExtraConfigStruct):
    """
    Stores additional arguments for quantization range initialization algorithms.

    :param data_loader: Provides an iterable over the given dataset.
    :param device: Device to perform initialization. If `device` is `None`
        then the device of the model parameters will be used.
    """

    def __init__(self, data_loader: NNCFDataLoader, device: Optional[str] = None):
        self._data_loader = data_loader
        self._device = device

    @property
    def data_loader(self) -> NNCFDataLoader:
        return self._data_loader

    @property
    def device(self) -> Optional[str]:
        return self._device

    @classmethod
    def get_id(cls) -> str:
        return "quantization_range_init_args"


@api()
class BNAdaptationInitArgs(NNCFExtraConfigStruct):
    """
    Stores additional arguments for batchnorm statistics adaptation algorithm.

    :param data_loader: Provides an iterable over the given dataset.
    :param device: Device to perform initialization. If `device` is `None`
        then the device of the model parameters will be used.
    """

    def __init__(self, data_loader: NNCFDataLoader, device: Optional[str] = None):
        self._data_loader = data_loader
        self._device = device

    @property
    def data_loader(self) -> NNCFDataLoader:
        return self._data_loader

    @property
    def device(self) -> Optional[str]:
        return self._device

    @classmethod
    def get_id(cls) -> str:
        return "bn_adaptation_init_args"


@api()
class ModelEvaluationArgs(NNCFExtraConfigStruct):
    """
    Stores additional arguments for running the model in the evaluation mode, should this be required for an algorithm.

    :param eval_fn: A function accepting a single argument - the model object - and returning the model's metric on
        the evaluation split of the dataset corresponding to the model.
    """

    def __init__(self, eval_fn: Callable[..., Any]):
        self.eval_fn = eval_fn

    @classmethod
    def get_id(cls) -> str:
        return "model_evaluation_args"



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import onnx

import nncf
from nncf.common.logging.logger import nncf_logger
from nncf.common.quantization.structs import QuantizationPreset
from nncf.data import Dataset
from nncf.onnx.graph.metatypes.groups import OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS
from nncf.onnx.graph.nncf_graph_builder import GraphConverter
from nncf.parameters import DropType
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedAccuracyRestorerParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.advanced_parameters import QuantizationParameters
from nncf.quantization.algorithms.accuracy_control.algorithm import QuantizationAccuracyRestorer
from nncf.quantization.algorithms.accuracy_control.algorithm import calculate_accuracy_drop
from nncf.quantization.algorithms.accuracy_control.evaluator import Evaluator
from nncf.quantization.algorithms.post_training.algorithm import PostTrainingQuantization
from nncf.quantization.quantize_model import quantize_with_tune_hyperparams
from nncf.quantization.quantize_model import warning_model_no_batchwise_support
from nncf.scopes import IgnoredScope

TTensor = TypeVar("TTensor")


def quantize_impl(
    model: onnx.ModelProto,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> onnx.ModelProto:
    """
    Implementation of the `quantize()` method for the ONNX backend.
    """
    if target_device == TargetDevice.CPU_SPR:
        raise nncf.ValidationError("target_device == CPU_SPR is not supported.")
    if mode is not None:
        raise ValueError(f"mode={mode} is not supported")
    if model.opset_import[0].version < 10:
        raise nncf.ValidationError("ONNX models with opset version < 10 do not support quantization.")
    if model.opset_import[0].version < 13:
        nncf_logger.warning(
            "ONNX models with 10 < opset version < 13 do not support per-channel quantization."
            " Per-tensor quantization will be applied."
        )
        if advanced_parameters is None:
            advanced_parameters = AdvancedQuantizationParameters()
        advanced_parameters.weights_quantization_params = QuantizationParameters(per_channel=False)
        advanced_parameters.activations_quantization_params = QuantizationParameters(per_channel=False)

    quantization_algorithm = PostTrainingQuantization(
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        ignored_scope=ignored_scope,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        advanced_parameters=advanced_parameters,
    )

    graph = GraphConverter.create_nncf_graph(model)
    warning_model_no_batchwise_support(graph, advanced_parameters, model_type, OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS)
    quantized_model = quantization_algorithm.apply(model, graph, dataset=calibration_dataset)

    return quantized_model


def quantize_with_accuracy_control_impl(
    model: onnx.ModelProto,
    calibration_dataset: Dataset,
    validation_dataset: Dataset,
    validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]],
    max_drop: float = 0.01,
    drop_type: DropType = DropType.ABSOLUTE,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters] = None,
    advanced_accuracy_restorer_parameters: Optional[AdvancedAccuracyRestorerParameters] = None,
) -> onnx.ModelProto:
    """
    Implementation of the `quantize_with_accuracy_control()` method for the ONNX backend.
    """
    if advanced_accuracy_restorer_parameters is None:
        advanced_accuracy_restorer_parameters = AdvancedAccuracyRestorerParameters()

    if advanced_quantization_parameters is None:
        advanced_quantization_parameters = AdvancedQuantizationParameters()

    quantized_model = quantize_impl(
        model=model,
        calibration_dataset=calibration_dataset,
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_quantization_parameters,
    )

    if advanced_accuracy_restorer_parameters.intermediate_model_dir:
        quantized_model_path = f"{advanced_accuracy_restorer_parameters.intermediate_model_dir}/intermediate_model.onnx"
        onnx.save(quantized_model, quantized_model_path)

    evaluator = Evaluator(validation_fn)
    evaluator.enable_iteration_count()

    initial_metric_results = evaluator.collect_metric_results(model, validation_dataset, model_name="initial")

    validation_dataset_size = evaluator.num_passed_iterations
    evaluator.disable_iteration_count()

    quantized_metric_results = evaluator.collect_metric_results(
        quantized_model, validation_dataset, model_name="quantized"
    )

    should_terminate, accuracy_drop = calculate_accuracy_drop(
        initial_metric_results.metric_value, quantized_metric_results.metric_value, max_drop, drop_type
    )

    nncf_logger.info(f"Accuracy drop: {accuracy_drop} ({drop_type})")

    if advanced_accuracy_restorer_parameters.tune_hyperparams and not should_terminate:
        tuned_quantized_model = quantize_with_tune_hyperparams(
            model,
            calibration_dataset,
            validation_dataset,
            validation_fn,
            initial_metric_results,
            quantized_metric_results,
            subset_size,
            preset,
            target_device,
            subset_size,
            fast_bias_correction,
            model_type,
            ignored_scope,
            advanced_quantization_parameters,
        )
        tuned_quantized_metric_results = evaluator.collect_metric_results(
            tuned_quantized_model, validation_dataset, model_name="tuned"
        )
        should_terminate, tuned_accuracy_drop = calculate_accuracy_drop(
            initial_metric_results.metric_value, tuned_quantized_metric_results.metric_value, max_drop, drop_type
        )

        nncf_logger.info(f"Accuracy drop (tuned): {tuned_accuracy_drop} ({drop_type})")

        if should_terminate or tuned_accuracy_drop < accuracy_drop:
            quantized_model = tuned_quantized_model
            quantized_metric_results = tuned_quantized_metric_results

    if not should_terminate:
        ranking_subset_size = subset_size
        if advanced_accuracy_restorer_parameters.ranking_subset_size is not None:
            ranking_subset_size = advanced_accuracy_restorer_parameters.ranking_subset_size

        accuracy_restorer = QuantizationAccuracyRestorer(
            ranking_subset_size,
            advanced_accuracy_restorer_parameters.max_num_iterations,
            max_drop,
            drop_type,
            advanced_accuracy_restorer_parameters.num_ranking_workers,
            advanced_accuracy_restorer_parameters.restore_mode,
        )
        quantized_model = accuracy_restorer.apply(
            model,
            initial_metric_results,
            quantized_model,
            quantized_metric_results,
            validation_dataset,
            validation_dataset_size,
            evaluator,
        )

    return quantized_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import tensorflow as tf

from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.layers.operation import NNCFOperation
from nncf.tensorflow.layers.wrapper import NNCFWrapper
from nncf.tensorflow.quantization.layers import FakeQuantize


def apply_overflow_fix(model: tf.keras.Model, op_names: List[str]) -> None:
    if not isinstance(model, tf.keras.Model):
        raise ValueError(f"Expected model to be a `tf.keras.Model` instance but got: {type(model)}")

    for wrapped_layer, weight_attr, op in get_nncf_operations(model, op_names):
        if op.half_range:
            apply_overflow_fix_to_layer(wrapped_layer, weight_attr, op)


def apply_overflow_fix_to_layer(wrapped_layer: NNCFWrapper, weight_attr: str, op: NNCFOperation) -> None:
    layer_weight = wrapped_layer.layer_weights[weight_attr]
    ops_weights = wrapped_layer.get_operation_weights(op.name)
    # Keep zero weights to prevent
    # zero quant calculation arithmetic errors
    mask = layer_weight == 0.0
    layer_weight_updated = op.call(layer_weight, ops_weights, False)

    # Assign exact zero to weights which
    # was exact zero before overflow fix
    layer_weight_updated = tf.where(mask, [0.0], layer_weight_updated)
    layer_weight.assign(layer_weight_updated)
    op.apply_overflow_fix(ops_weights)


def collect_fake_quantize_layers(model: tf.keras.Model) -> List[FakeQuantize]:
    """
    Collects all fake quantize layers from the provided model.

    :param model: An instance of the `tf.keras.Model` class.
    :return: A list of fake quantize layers.
    """
    fq_layers = []
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            fq_layers.extend(collect_fake_quantize_layers(layer))
        if isinstance(layer, FakeQuantize):
            fq_layers.append(layer)
    return fq_layers



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.tensor.functions import io as io
from nncf.tensor.functions import linalg as linalg
from nncf.tensor.functions.numeric import abs as abs
from nncf.tensor.functions.numeric import all as all
from nncf.tensor.functions.numeric import allclose as allclose
from nncf.tensor.functions.numeric import any as any
from nncf.tensor.functions.numeric import arange as arange
from nncf.tensor.functions.numeric import argsort as argsort
from nncf.tensor.functions.numeric import as_tensor_like as as_tensor_like
from nncf.tensor.functions.numeric import astype as astype
from nncf.tensor.functions.numeric import ceil as ceil
from nncf.tensor.functions.numeric import clip as clip
from nncf.tensor.functions.numeric import concatenate as concatenate
from nncf.tensor.functions.numeric import count_nonzero as count_nonzero
from nncf.tensor.functions.numeric import device as device
from nncf.tensor.functions.numeric import diag as diag
from nncf.tensor.functions.numeric import dtype as dtype
from nncf.tensor.functions.numeric import expand_dims as expand_dims
from nncf.tensor.functions.numeric import eye as eye
from nncf.tensor.functions.numeric import finfo as finfo
from nncf.tensor.functions.numeric import flatten as flatten
from nncf.tensor.functions.numeric import from_numpy as from_numpy
from nncf.tensor.functions.numeric import isclose as isclose
from nncf.tensor.functions.numeric import isempty as isempty
from nncf.tensor.functions.numeric import item as item
from nncf.tensor.functions.numeric import log2 as log2
from nncf.tensor.functions.numeric import logical_or as logical_or
from nncf.tensor.functions.numeric import masked_mean as masked_mean
from nncf.tensor.functions.numeric import masked_median as masked_median
from nncf.tensor.functions.numeric import matmul as matmul
from nncf.tensor.functions.numeric import max as max
from nncf.tensor.functions.numeric import maximum as maximum
from nncf.tensor.functions.numeric import mean as mean
from nncf.tensor.functions.numeric import median as median
from nncf.tensor.functions.numeric import min as min
from nncf.tensor.functions.numeric import minimum as minimum
from nncf.tensor.functions.numeric import moveaxis as moveaxis
from nncf.tensor.functions.numeric import multiply as multiply
from nncf.tensor.functions.numeric import ones_like as ones_like
from nncf.tensor.functions.numeric import percentile as percentile
from nncf.tensor.functions.numeric import power as power
from nncf.tensor.functions.numeric import quantile as quantile
from nncf.tensor.functions.numeric import reshape as reshape
from nncf.tensor.functions.numeric import round as round
from nncf.tensor.functions.numeric import searchsorted as searchsorted
from nncf.tensor.functions.numeric import squeeze as squeeze
from nncf.tensor.functions.numeric import stack as stack
from nncf.tensor.functions.numeric import sum as sum
from nncf.tensor.functions.numeric import tensor as tensor
from nncf.tensor.functions.numeric import transpose as transpose
from nncf.tensor.functions.numeric import unsqueeze as unsqueeze
from nncf.tensor.functions.numeric import unstack as unstack
from nncf.tensor.functions.numeric import var as var
from nncf.tensor.functions.numeric import where as where
from nncf.tensor.functions.numeric import zeros as zeros
from nncf.tensor.functions.numeric import zeros_like as zeros_like


def _initialize_backends():
    import contextlib

    import nncf.tensor.functions.numpy_io
    import nncf.tensor.functions.numpy_linalg
    import nncf.tensor.functions.numpy_numeric

    with contextlib.suppress(ImportError):
        import nncf.tensor.functions.torch_io
        import nncf.tensor.functions.torch_linalg
        import nncf.tensor.functions.torch_numeric  # noqa F401

    with contextlib.suppress(ImportError):
        import nncf.tensor.functions.ov  # noqa: F401


_initialize_backends()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

from nncf.common.graph.layer_attributes import BaseLayerAttributes


class OVLayerAttributes(BaseLayerAttributes):
    """
    This class stores additional information about nodes that needs to be processed during compression.
    """

    def __init__(
        self,
        constant_attributes: Dict[int, Any],
        layer_attributes: Optional[BaseLayerAttributes] = None,
        inputs_attributes: Optional[Dict[Any, Any]] = None,
    ):
        """
        :param constant_attributes: Map of weights port ID to corresponding const attributes.
        :param layer_attributes: Map of weights port ID to corresponding common layer attributes.
        :param inputs_attributes: Activation attributes.
        """
        self._constant_attributes = constant_attributes
        self._layer_attributes = layer_attributes
        self._inputs_attributes = inputs_attributes

    @property
    def constant_attributes(self) -> Dict[int, Any]:
        return self._constant_attributes

    @property
    def layer_attributes(self) -> Optional[BaseLayerAttributes]:
        return self._layer_attributes

    @property
    def input_attributes(self) -> Optional[Dict[Any, Any]]:
        return self._inputs_attributes

    def get_const_port_ids(self) -> List[int]:
        """
        Returns indices of input ports corresponding to the constant nodes.

        :returns: List of input port indices with constants.
        """
        if self._constant_attributes is not None:
            return list(self._constant_attributes.keys())
        return []



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import operator
from typing import Any, Dict, Optional, Tuple, TypeVar, Union

import nncf
from nncf.common.utils.backend import BackendType
from nncf.tensor.definitions import TensorBackend
from nncf.tensor.definitions import TensorDataType
from nncf.tensor.definitions import TensorDeviceType

TTensor = TypeVar("TTensor")


class Tensor:
    """
    An interface to framework specific tensors for common NNCF algorithms.
    """

    def __init__(self, data: Optional[TTensor]):
        self._data = data.data if isinstance(data, Tensor) else data

    @property
    def data(self) -> TTensor:
        return self._data

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(self.data.shape)

    @property
    def ndim(self) -> int:
        return self.data.ndim

    @property
    def device(self) -> TensorDeviceType:
        return _call_function("device", self)

    @property
    def dtype(self) -> TensorDataType:
        return _call_function("dtype", self)

    @property
    def backend(self) -> TensorBackend:
        return _call_function("backend", self)

    @property
    def size(self) -> int:
        return _call_function("size", self)

    def __bool__(self) -> bool:
        return bool(self.data)

    def __iter__(self):
        return TensorIterator(self.data)

    def __getitem__(self, index: Union[Tensor, int, Tuple[Union[Tensor, int], ...]]) -> Tensor:
        return Tensor(self.data[unwrap_index(index)])

    def __setitem__(self, index: Union[Tensor, int, Tuple[Union[Tensor, int], ...]], value: Any) -> None:
        self.data[unwrap_index(index)] = unwrap_tensor_data(value)

    def __str__(self) -> str:
        return f"nncf.Tensor({str(self.data)})"

    def __repr__(self) -> str:
        return f"nncf.Tensor({repr(self.data)})"

    # built-in operations

    def __add__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data + unwrap_tensor_data(other))

    def __radd__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) + self.data)

    def __iadd__(self, other: Union[Tensor, float]) -> Tensor:
        self._data += unwrap_tensor_data(other)
        return self

    def __sub__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data - unwrap_tensor_data(other))

    def __rsub__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) - self.data)

    def __isub__(self, other: Union[Tensor, float]) -> Tensor:
        self._data -= unwrap_tensor_data(other)
        return self

    def __mul__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data * unwrap_tensor_data(other))

    def __rmul__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) * self.data)

    def __imul__(self, other: Union[Tensor, float]) -> Tensor:
        self._data *= unwrap_tensor_data(other)
        return self

    def __pow__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data ** unwrap_tensor_data(other))

    def __rpow__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(unwrap_tensor_data(other) ** self.data)

    def __ipow__(self, other: Union[Tensor, float]) -> Tensor:
        self._data **= unwrap_tensor_data(other)
        return self

    def __truediv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_op_nowarn", self, other, operator.truediv)

    def __rtruediv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_reverse_op_nowarn", self, other, operator.truediv)

    def __itruediv__(self, other: Union[Tensor, float]) -> Tensor:
        self._data /= unwrap_tensor_data(other)
        return self

    def __floordiv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_op_nowarn", self, other, operator.floordiv)

    def __rfloordiv__(self, other: Union[Tensor, float]) -> Tensor:
        return _call_function("_binary_reverse_op_nowarn", self, other, operator.floordiv)

    def __ifloordiv__(self, other: Union[Tensor, float]) -> Tensor:
        self._data /= unwrap_tensor_data(other)
        return self

    def __matmul__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data @ unwrap_tensor_data(other))

    def __neg__(self) -> Tensor:
        return Tensor(-self.data)

    # Comparison operators

    def __lt__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data < unwrap_tensor_data(other))

    def __le__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data <= unwrap_tensor_data(other))

    def __eq__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data == unwrap_tensor_data(other))

    def __ne__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data != unwrap_tensor_data(other))

    def __gt__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data > unwrap_tensor_data(other))

    def __ge__(self, other: Union[Tensor, float]) -> Tensor:
        return Tensor(self.data >= unwrap_tensor_data(other))

    # Tensor functions

    def squeeze(self, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
        return _call_function("squeeze", self, axis)

    def flatten(self) -> Tensor:
        return _call_function("flatten", self)

    def max(self, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: Optional[bool] = False) -> Tensor:
        return _call_function("max", self, axis, keepdims)

    def min(self, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: Optional[bool] = False) -> Tensor:
        return _call_function("min", self, axis, keepdims)

    def abs(self) -> Tensor:
        return _call_function("abs", self)

    def isempty(self) -> bool:
        return _call_function("isempty", self)

    def astype(self, dtype: TensorDataType) -> Tensor:
        return _call_function("astype", self, dtype)

    def reshape(self, shape: Tuple[int, ...]) -> Tensor:
        return _call_function("reshape", self, shape)

    def item(self) -> float:
        return _call_function("item", self)

    def clone(self) -> float:
        return _call_function("clone", self)

    def as_numpy_tensor(self) -> Tensor:
        return _call_function("as_numpy_tensor", self)


def _call_function(func_name: str, *args):
    """
    Call function from functions.py to avoid circular imports.

    :param func_name: Name of function.
    :return: Result of function call.
    """
    from nncf.tensor.functions import numeric

    fn = getattr(numeric, func_name)
    return fn(*args)


class TensorIterator:
    """Iterator for Tensor class"""

    def __init__(self, tensor):
        self._tensor = tensor
        self._index = 0

    def __next__(self) -> Tensor:
        if self._index < len(self._tensor):
            result = self._tensor[self._index]
            self._index += 1
            return Tensor(result)

        raise StopIteration


def unwrap_index(obj: Union[Any, Tuple[Any, ...]]) -> Union[TTensor, Tuple[TTensor, ...]]:
    """
    Unwraps the tensor data from the input object or tuple of objects.

    :param obj: The object to unwrap.
    :return: The unwrapped tensor data or tuple of unwrapped tensor data.
    """
    if isinstance(obj, tuple):
        return tuple(unwrap_tensor_data(o) for o in obj)
    return unwrap_tensor_data(obj)


def unwrap_tensor_data(obj: Any) -> TTensor:
    """
    Return the data of a Tensor object, or the object itself if it is not a Tensor.

    :param obj: The object to unwrap.
    :return: The data of the Tensor object, or the object itself.
    """
    return obj.data if isinstance(obj, Tensor) else obj


def get_tensor_backend(backend: BackendType) -> TensorBackend:
    """
    Returns a tensor backend based on the provided backend.

    :param backend: Backend type.
    :return: Corresponding tensor backend type.
    """
    BACKEND_TO_TENSOR_BACKEND: Dict[BackendType, TensorBackend] = {
        BackendType.OPENVINO: TensorBackend.numpy,
        BackendType.ONNX: TensorBackend.numpy,
        BackendType.TORCH_FX: TensorBackend.torch,
        BackendType.TORCH: TensorBackend.torch,
    }
    if backend not in BACKEND_TO_TENSOR_BACKEND:
        raise nncf.ValidationError(f"Unsupported backend type: {backend}")

    return BACKEND_TO_TENSOR_BACKEND[backend]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
from operator import mul
from typing import Tuple

from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.tensor import Tensor
from nncf.tensor import functions as fns


def process_stats(stats: WCTensorStatistic, subset_size: int) -> Tuple[Tensor, Tensor]:
    """
    A function for processing activations. Shared between AWQ, Scale Estimation and LoRA Correction algorithms.

    :param stats: An object containing statistics for the layer.
    :param subset_size: The number of samples for AWQ.
    :return: tuple of the following tensors:
        s - maximum channel magnitude across samples [HiddenDim]
        X - average channel magnitude across tokens in the sequence [HiddenDim, min(SampleSize, ~subset_size)]
    """
    X = fns.stack(stats.mean_values)  # [SampleSize, HiddenDim]
    X_full = fns.transpose(X)  # [HiddenDim, SampleSize]

    # prevent high memory and time consumption
    if X_full.shape[1] > subset_size:
        # activations were reduced across all but the last dimension
        lens = [reduce(mul, shape[:-1], 1) for shape in stats.shape_values]
        step = X_full.shape[1] // subset_size
        idxs = [i[0] for i in sorted(enumerate(lens), key=lambda x: -x[1])][::step]
        X = X_full[:, idxs]  # [HiddenDim, ~SubsetSize]
    else:
        X = X_full
    s = fns.max(fns.abs(X_full), axis=1)  # [HiddenDim]
    return s, X



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Neural Network Compression Framework (NNCF) for enhanced OpenVINO™ inference.
"""

from nncf.common.logging import nncf_logger as nncf_logger
from nncf.common.logging.logger import disable_logging as disable_logging
from nncf.common.logging.logger import set_log_level as set_log_level
from nncf.common.strip import strip as strip
from nncf.config import NNCFConfig as NNCFConfig
from nncf.data import Dataset as Dataset
from nncf.errors import BufferFullError as BufferFullError
from nncf.errors import InstallationError as InstallationError
from nncf.errors import InternalError as InternalError
from nncf.errors import InvalidCollectorTypeError as InvalidCollectorTypeError
from nncf.errors import InvalidPathError as InvalidPathError
from nncf.errors import InvalidQuantizerGroupError as InvalidQuantizerGroupError
from nncf.errors import ModuleNotFoundError as ModuleNotFoundError
from nncf.errors import ParameterNotSupportedError as ParameterNotSupportedError
from nncf.errors import StatisticsCacheError as StatisticsCacheError
from nncf.errors import UnknownDatasetError as UnknownDatasetError
from nncf.errors import UnsupportedBackendError as UnsupportedBackendError
from nncf.errors import UnsupportedDatasetError as UnsupportedDatasetError
from nncf.errors import UnsupportedModelError as UnsupportedModelError
from nncf.errors import UnsupportedVersionError as UnsupportedVersionError
from nncf.errors import ValidationError as ValidationError
from nncf.parameters import BackupMode as BackupMode
from nncf.parameters import CompressWeightsMode as CompressWeightsMode
from nncf.parameters import DropType as DropType
from nncf.parameters import ModelType as ModelType
from nncf.parameters import QuantizationMode as QuantizationMode
from nncf.parameters import SensitivityMetric as SensitivityMetric
from nncf.parameters import TargetDevice as TargetDevice
from nncf.quantization import QuantizationPreset as QuantizationPreset
from nncf.quantization import compress_weights as compress_weights
from nncf.quantization import quantize as quantize
from nncf.quantization import quantize_with_accuracy_control as quantize_with_accuracy_control
from nncf.quantization.advanced_parameters import (
    AdvancedAccuracyRestorerParameters as AdvancedAccuracyRestorerParameters,
)
from nncf.quantization.advanced_parameters import AdvancedAWQParameters as AdvancedAWQParameters
from nncf.quantization.advanced_parameters import AdvancedBiasCorrectionParameters as AdvancedBiasCorrectionParameters
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters as AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import AdvancedGPTQParameters as AdvancedGPTQParameters
from nncf.quantization.advanced_parameters import AdvancedLoraCorrectionParameters as AdvancedLoraCorrectionParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters as AdvancedQuantizationParameters
from nncf.quantization.advanced_parameters import AdvancedScaleEstimationParameters as AdvancedScaleEstimationParameters
from nncf.quantization.advanced_parameters import AdvancedSmoothQuantParameters as AdvancedSmoothQuantParameters
from nncf.quantization.advanced_parameters import OverflowFix as OverflowFix
from nncf.scopes import IgnoredScope as IgnoredScope
from nncf.scopes import Subgraph as Subgraph
from nncf.version import __version__ as __version__

_SUPPORTED_FRAMEWORKS = ["torch", "tensorflow", "onnx", "openvino"]


from importlib.util import find_spec as _find_spec  # noqa: E402
from pathlib import Path as _Path  # noqa: E402

_AVAILABLE_FRAMEWORKS = {}

for fw_name in _SUPPORTED_FRAMEWORKS:
    spec = _find_spec(fw_name)
    framework_present = False
    if spec is not None and spec.origin is not None:
        origin_path = _Path(spec.origin)
        here = _Path(__file__)
        if origin_path not in here.parents:
            # if the framework is not present, spec may still be not None because
            # it found our nncf.*backend_name* subpackage, and spec.origin will point to a folder in NNCF code
            framework_present = True
    _AVAILABLE_FRAMEWORKS[fw_name] = framework_present

if not any(_AVAILABLE_FRAMEWORKS.values()):
    nncf_logger.error(
        "Neither PyTorch, TensorFlow, ONNX or OpenVINO Python packages have been found in your Python "
        "environment.\n"
        "Please install one of the supported frameworks above in order to use NNCF on top of it.\n"
        "See the installation guide at https://github.com/openvinotoolkit/nncf#installation-guide for help."
    )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Set, Tuple

import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.openvino.graph.metatypes.groups import FAKE_QUANTIZE_OPERATIONS
from nncf.openvino.graph.model_utils import remove_fq_from_inputs
from nncf.openvino.graph.node_utils import get_bias_value
from nncf.openvino.graph.node_utils import get_parameter_node_name
from nncf.openvino.graph.node_utils import get_result_node_name
from nncf.openvino.graph.node_utils import is_node_with_bias
from nncf.openvino.graph.transformations.command_creation import OVCommandCreator
from nncf.openvino.graph.transformations.commands import OVBiasCorrectionCommand
from nncf.openvino.graph.transformations.commands import OVModelExtractionCommand
from nncf.openvino.graph.transformations.commands import OVOutputInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.statistics.collectors import get_mean_statistic_collector
from nncf.openvino.statistics.collectors import get_raw_stat_collector
from nncf.quantization.algorithms.bias_correction.backend import BiasCorrectionAlgoBackend
from nncf.tensor import Tensor


class OVBiasCorrectionAlgoBackend(BiasCorrectionAlgoBackend):

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> OVBiasCorrectionCommand:
        return OVCommandCreator.create_command_to_update_bias(node, bias_value.data, nncf_graph)

    @staticmethod
    def model_extraction_command(
        input_ids: Set[Tuple[str, int]], output_ids: Set[Tuple[str, int]]
    ) -> OVModelExtractionCommand:
        return OVModelExtractionCommand(input_ids, output_ids)

    @staticmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: OVTargetPoint) -> OVOutputInsertionCommand:
        return OVOutputInsertionCommand(target_point)

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size, inplace)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: str) -> Tensor:
        return Tensor(raw_data[output_name])

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        constant_ports = node.layer_attributes.get_const_port_ids()
        activation_ports = [
            e.input_port_id for e in nncf_graph.get_input_edges(node) if e.input_port_id not in constant_ports
        ]
        assert len(activation_ports) == 1
        return activation_ports[0]

    @staticmethod
    def get_bias_value(node: NNCFNode, model: ov.Model, nncf_graph: NNCFGraph) -> Tensor:
        return Tensor(get_bias_value(node, nncf_graph, model))

    @staticmethod
    def get_input_name(model: ov.Model, node_name: str, input_port_id: int) -> str:
        return get_parameter_node_name(node_name, input_port_id)

    @staticmethod
    def get_output_name(model: ov.Model, node_name: str, output_port_id: int) -> str:
        return get_result_node_name(node_name, output_port_id)

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        if node.layer_attributes is None:
            return False
        const_port_ids = node.layer_attributes.get_const_port_ids()
        assert len(const_port_ids) == 1
        weight_node = nncf_graph.get_input_edge_by_port_id(node, const_port_ids[0]).from_node
        return weight_node.metatype in FAKE_QUANTIZE_OPERATIONS

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def remove_fq_from_inputs(model: ov.Model, nncf_graph: NNCFGraph) -> ov.Model:
        return remove_fq_from_inputs(model, nncf_graph)

    @staticmethod
    def get_port_id(target_point: OVTargetPoint) -> int:
        return target_point.port_id



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Tuple, TypeVar

from nncf.common.factory import CommandCreatorFactory
from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.quantization.advanced_parameters import RestoreMode

TModel = TypeVar("TModel")


def find_quantizer_nodes_to_cut(
    graph: NNCFGraph,
    quantizer_node: NNCFNode,
    quantizer_metatypes: List[OperatorMetatype],
    const_metatypes: List[OperatorMetatype],
    quantizable_metatypes: List[OperatorMetatype],
    quantize_agnostic_metatypes: List[OperatorMetatype],
) -> Tuple[List[NNCFNode], List[NNCFNode]]:
    """
    Finds quantizer nodes that should be removed in addition to `quantizer_node` to get
    the correct model for inference. Returns the list of quantizer nodes (`quantizer_node` + nodes
    which were found) and the list of nodes that will be reverted to original precision if
    quantizer nodes are removed.

    :param graph: The NNCFGraph without shapeof subgraphs.
    :param quantizer_node: The quantizer node that we want to remove.
    :param quantizer_metatypes: List of quantizer metatypes.
    :param const_metatypes: List of constant metatypes.
    :param quantizable_metatypes: List of metatypes for operations
        that may be quantized.
    :param quantize_agnostic_metatypes: List of quantize agnostic metatypes.
    :return: A tuple (quantizer_nodes, ops) where
        - `quantizer_nodes` is the list of quantizer nodes
        - `ops` is the list of nodes that will be reverted to original precision
        if `quantizer_nodes` are removed.
    """

    def _parse_node_relatives(node: NNCFNode, is_parents: bool):
        relatives = graph.get_previous_nodes(node) if is_parents else graph.get_next_nodes(node)
        for relative in relatives:
            if relative.metatype in quantizable_metatypes:
                if is_parents:
                    if relative in seen_children:
                        continue
                    to_see_children.append(relative)
                else:
                    ops_to_return_in_orig_prec.add(relative)
                    if relative not in seen_parents:
                        to_see_parents.append(relative)
            elif relative.metatype in quantizer_metatypes:
                if is_parents:
                    if relative in seen_children:
                        continue
                    if relative not in to_cut:
                        to_cut.append(relative)
                    to_see_children.append(relative)
                    # We should see parents for the `relative` node here only if they are
                    # all quantizers. This covers the quantize-dequantize case, where we
                    # should see parents for the dequantize node.
                    if all(x.metatype in quantizer_metatypes for x in graph.get_previous_nodes(relative)):
                        to_see_parents.append(relative)
                elif node.metatype in quantizer_metatypes:
                    # `node` is a quantizer (quantize-dequantize case) here, and `relative`
                    # is the dequantizer. So, we should cut `relative` and look at its children.
                    if relative not in to_cut:
                        to_cut.append(relative)
                    to_see_children.append(relative)
                else:
                    seen_children.append(relative)
            elif relative.metatype not in const_metatypes:
                if relative not in seen_parents:
                    to_see_parents.append(relative)
                if relative not in seen_children and relative.metatype in quantize_agnostic_metatypes:
                    to_see_children.append(relative)

        seen_list = seen_parents if is_parents else seen_children
        seen_list.append(node)

    seen_children, seen_parents = [], []
    to_see_children, to_see_parents = [quantizer_node], []
    to_cut = [quantizer_node]
    ops_to_return_in_orig_prec = set()

    while to_see_parents or to_see_children:
        if to_see_children:
            _parse_node_relatives(to_see_children.pop(), is_parents=False)
        if to_see_parents:
            _parse_node_relatives(to_see_parents.pop(), is_parents=True)

    return to_cut, list(ops_to_return_in_orig_prec)


def revert_operations_to_floating_point_precision(
    operations: List[NNCFNode],
    quantizers: List[NNCFNode],
    quantized_model: TModel,
    quantized_model_graph: NNCFGraph,
    restore_mode: RestoreMode,
    op_with_weights_metatypes: List[OperatorMetatype],
    is_node_with_weight_fn: Callable[[NNCFNode], bool],
    get_weight_tensor_port_ids_fn: Callable[[NNCFNode], List[int]],
) -> TModel:
    """
    Reverts provided operations to floating-point precision by removing
    quantizers. Restores original bias for operations with bias.
    Restores original weights for operations with weights.

    :param operations: List of operations to revert in floating-point precision.
    :param quantizers: List of quantizers that should be removed to revert
        operations to floating-point precision.
    :param quantized_model: Quantized model in which provided operations
        should be reverted to floating-point precision.
    :param quantized_model_graph: The graph which was built for `quantized_model`.
    :param restore_mode: Restore mode.
    :param op_with_weights_metatypes: List of operation metatypes that can be reverted to representation
        with int8 weights.
    :param is_node_with_weight_fn: Checks if the node has a weight or not. Returns `True` if `node` corresponds
        to the operation with weights, `False` otherwise.
    :param get_weight_tensor_port_ids_fn: Returns node's input port indices with weights tensors.
    :return: The model where `operations` were reverted to floating-point precision.
    """
    transformation_layout = TransformationLayout()

    command_creator = CommandCreatorFactory.create(quantized_model)

    should_remove_fq = {}
    should_revert_weights_to_fp32 = {}
    if restore_mode == RestoreMode.ONLY_ACTIVATIONS:
        for node in operations:
            if (
                node.metatype in op_with_weights_metatypes
                and node.layer_attributes
                and node.layer_attributes.constant_attributes
            ):
                input_edges = quantized_model_graph.get_input_edges(node)
                for port_id in node.layer_attributes.get_const_port_ids():
                    fq_weight = input_edges[port_id].from_node
                    should_remove_fq[fq_weight.node_name] = False
                    should_revert_weights_to_fp32[node.node_name] = False

    for node in quantizers:
        if should_remove_fq.get(node.node_name, True):
            transformation_layout.register(command_creator.create_command_to_remove_quantizer(node))

    for node in operations:
        original_bias = node.attributes.get("original_bias", None)
        if original_bias is not None:
            transformation_layout.register(
                command_creator.create_command_to_update_bias(node, original_bias, quantized_model_graph)
            )

        if not should_revert_weights_to_fp32.get(node.node_name, True):
            continue

        if is_node_with_weight_fn(node):
            weight_port_ids = get_weight_tensor_port_ids_fn(node)
            for port_id in weight_port_ids:
                original_weight = node.attributes.get(f"original_weight.{port_id}", None)
                if original_weight is not None:
                    transformation_layout.register(
                        command_creator.create_command_to_update_weight(node, original_weight, port_id)
                    )

    model_transformer = ModelTransformerFactory.create(quantized_model)
    transformed_model = model_transformer.transform(transformation_layout)

    return transformed_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from functools import lru_cache
from typing import cast


class NNCFLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        super().__init__(name, level)

    @lru_cache(None)
    def _log_once(self, level: int, msg: str) -> None:
        self.log(level, msg)

    def debug_once(self, msg: str) -> None:
        """
        Log a message at the DEBUG level, ensuring the message is logged only once.
        """
        self._log_once(logging.DEBUG, msg)

    def info_once(self, msg: str) -> None:
        """
        Log a message at the INFO level, ensuring the message is logged only once.
        """
        self._log_once(logging.INFO, msg)

    def warning_once(self, msg: str) -> None:
        """
        Log a message at the WARNING level, ensuring the message is logged only once.
        """
        self._log_once(logging.WARNING, msg)


NNCF_LOGGER_NAME = "nncf"

logging.setLoggerClass(NNCFLogger)
nncf_logger = cast(NNCFLogger, logging.getLogger(NNCF_LOGGER_NAME))
nncf_logger.propagate = False

stdout_handler = logging.StreamHandler(sys.stdout)
fmt_string = "%(levelname)s:%(name)s:%(message)s"
fmt = logging.Formatter(fmt_string)
fmt_no_newline = logging.Formatter(
    fmt_string,
)
stdout_handler.setFormatter(fmt)
stdout_handler.setLevel(logging.INFO)
nncf_logger.addHandler(stdout_handler)
nncf_logger.setLevel(logging.INFO)


def set_log_level(level: int) -> None:
    """
    Sets the log level for the NNCF logging.
    :param level: An integer passed into the underlying `logging.Logger` object,
      i.e. the regular `logging` log levels must be used here such as `logging.WARNING`
      or `logging.DEBUG`.
    """
    nncf_logger.setLevel(level)
    for handler in nncf_logger.handlers:
        handler.setLevel(level)


def set_log_file(filename: str) -> None:
    """
    Sets the log file for the NNCF logging.

    :param filename: Path to the file to save the log.
    """
    file_handler = logging.FileHandler(filename, encoding="utf-8")
    file_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
    nncf_logger.addHandler(file_handler)


def disable_logging() -> None:
    """
    Disables NNCF logging entirely. `FutureWarning`s are still shown.
    """
    nncf_logger.handlers = []


NNCFDeprecationWarning = FutureWarning


def warn_bkc_version_mismatch(backend: str, bkc_version: str, current_version: str) -> None:
    nncf_logger.warning(
        f"NNCF provides best results with {backend}{bkc_version}, "
        f"while current {backend} version is {current_version}. "
        f"If you encounter issues, consider switching to {backend}{bkc_version}"
    )
